import logging

from django.contrib.auth import views as auth_views

logger = logging.getLogger(__name__)


class MfaLoginView(auth_views.LoginView):
    """
    View to extend the default login behaviour from :class:`~django.contrib.auth.views.LoginView` with
    multi-factor-authentication.
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "authentication/login_mfa.html"



"""
This module contains view actions related to the imprint.
"""
import logging

from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.shortcuts import redirect
from django.utils.translation import ugettext as _
from django.http import HttpResponseNotFound
from django.views.decorators.http import require_POST

from ...decorators import permission_required
from ...models import ImprintPage, ImprintPageTranslation, ImprintPageFeedback

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.delete_imprintpage")
def delete_imprint(request, region_slug):
    """
    Delete imprint object

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param region_slug: The slug of the current region
    :type region_slug: str

    :raises ~django.http.Http404: If no imprint exists for the region

    :return: A redirection to the :class:`~integreat_cms.cms.views.imprint.imprint_form_view.ImprintFormView`
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = request.region

    try:
        imprint = region.imprint
    except ImprintPage.DoesNotExist as e:
        raise Http404 from e

    logger.debug("%r deleted by %r", imprint, request.user)

    imprint.delete()
    ImprintPageFeedback.objects.filter(region=region).delete()
    messages.success(request, _("Imprint was successfully deleted"))

    return redirect(
        "edit_imprint",
        **{
            "region_slug": region_slug,
        },
    )


def expand_imprint_translation_id(request, imprint_translation_id):
    """
    Searches for an imprint translation with corresponding ID and redirects browser to web app

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param imprint_translation_id: The id of the requested imprint translation
    :type imprint_translation_id: int

    :return: A redirection to :class:`~integreat_cms.core.settings.WEBAPP_URL`
    :rtype: ~django.http.HttpResponseRedirect
    """

    imprint_translation = ImprintPageTranslation.objects.get(
        id=imprint_translation_id
    ).public_version

    if imprint_translation and not imprint_translation.page.archived:
        return redirect(settings.WEBAPP_URL + imprint_translation.get_absolute_url())
    return HttpResponseNotFound("<h1>Imprint not found</h1>")



from django.conf import settings
from django.urls import include, path
from django.views.static import serve

from integreat_cms.core.urls import urlpatterns

#: Add the debug serve view to test the download of XLIFF files
urlpatterns += [
    path(
        "",
        include(
            (
                [
                    path(
                        f"{settings.XLIFF_URL}<path:path>".lstrip("/"),
                        serve,
                        {"document_root": settings.XLIFF_DOWNLOAD_DIR},
                    )
                ],
                "xliff_files",
            )
        ),
    ),
]



from django import forms
from ..models import MediaFile


class IconWidget(forms.HiddenInput):
    """
    A custom widget to render the icon field
    """

    #: The template to use for this widget
    template_name = "icon_widget.html"

    def get_context(self, name, value, attrs):
        """
        This function gets the context of icon fields

        :param name: the supplied name
        :type name: str

        :param value: the supplied values
        :type value: str

        :param attrs: the supplied attrs
        :type attrs: list

        :return: context
        :rtype: dict
        """

        context = super().get_context(name, value, attrs)
        if value:
            context["widget"]["document"] = MediaFile.objects.get(id=value)

        return context



import logging

from django.contrib import messages
from django.contrib.auth import update_session_auth_hash
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView
from django.shortcuts import render, redirect
from django.views.decorators.cache import never_cache

from ...forms import UserEmailForm, UserPasswordForm

logger = logging.getLogger(__name__)


class UserSettingsView(TemplateView):
    """
    View for the individual user settings
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "settings/user_settings.html"

    @never_cache
    def get(self, request, *args, **kwargs):
        r"""
        Render :class:`~integreat_cms.cms.forms.users.user_email_form.UserEmailForm` and
        :class:`~integreat_cms.cms.forms.users.user_password_form.UserPasswordForm`

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        user = request.user
        user_email_form = UserEmailForm(instance=user)
        user_password_form = UserPasswordForm(instance=user)

        return render(
            request,
            self.template_name,
            {
                "keys": user.mfa_keys.all(),
                "user_email_form": user_email_form,
                "user_password_form": user_password_form,
            },
        )

    # pylint: disable=unused-argument, too-many-branches
    def post(self, request, *args, **kwargs):
        r"""
        Submit :class:`~integreat_cms.cms.forms.users.user_email_form.UserEmailForm` and
        :class:`~integreat_cms.cms.forms.users.user_password_form.UserPasswordForm` and save :class:`~django.contrib.auth.models.User`
        object

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """
        region = request.region

        user = request.user

        if request.POST.get("submit_form") == "email_form":
            user_email_form = UserEmailForm(data=request.POST, instance=user)
            if not user_email_form.is_valid():

                # Add error messages
                for field in user_email_form:
                    for error in field.errors:
                        messages.error(request, _(error))
                for error in user_email_form.non_field_errors():
                    messages.error(request, _(error))

            elif not user_email_form.has_changed():
                messages.info(request, _("No changes made"))
            else:
                user_email_form.save()
                messages.success(request, _("E-mail-address was successfully saved"))

        elif request.POST.get("submit_form") == "password_form":
            user_password_form = UserPasswordForm(data=request.POST, instance=user)
            if not user_password_form.is_valid():

                # Add error messages
                for field in user_password_form:
                    for error in field.errors:
                        messages.error(request, _(error))
                for error in user_password_form.non_field_errors():
                    messages.error(request, _(error))

            elif not user_password_form.has_changed():
                messages.info(request, _("No changes made"))
            else:
                user = user_password_form.save()
                # Prevent user from being logged out after password has changed
                update_session_auth_hash(request, user)
                messages.success(request, _("Password was successfully saved"))

        kwargs = {"region_slug": region.slug} if region else {}
        return redirect("user_settings", **kwargs)



"""
This module contains action methods for events (archive, restore, ...)
"""
import json
import logging

from django.contrib import messages
from django.db.models import Subquery, OuterRef
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from linkcheck.models import Link

from ...constants import status
from ...decorators import permission_required
from ...models import Region, POITranslation

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.change_event")
def archive(request, event_id, region_slug, language_slug):
    """
    Set archived flag for an event

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param event_id: internal id of the event to be archived
    :type event_id: int

    :param region_slug: slug of the region which the event belongs to
    :type region_slug: str

    :param language_slug: current GUI language slug
    :type language_slug: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    region = request.region
    event = get_object_or_404(region.events, id=event_id)

    event.archived = True
    event.save()

    # Delete related link objects as they are no longer required
    Link.objects.filter(event_translation__event=event).delete()

    logger.debug("%r archived by %r", event, request.user)
    messages.success(request, _("Event was successfully archived"))

    return redirect(
        "events",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@require_POST
@permission_required("cms.change_event")
def copy(request, event_id, region_slug, language_slug):
    """
    Duplicates the given event and all of its translations.

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param event_id: internal id of the event to be copied
    :type event_id: int

    :param region_slug: slug of the region which the event belongs to
    :type region_slug: str

    :param language_slug: current GUI language slug
    :type language_slug: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    region = request.region
    event = get_object_or_404(region.events, id=event_id)

    event.copy(request.user)

    logger.debug("%r copied by %r", event, request.user)
    messages.success(request, _("Event was successfully copied"))

    return redirect(
        "events", **{"region_slug": region_slug, "language_slug": language_slug}
    )


@require_POST
@permission_required("cms.change_event")
def restore(request, event_id, region_slug, language_slug):
    """
    Remove archived flag for an event

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param event_id: internal id of the event to be un-archived
    :type event_id: int

    :param region_slug: slug of the region which the event belongs to
    :type region_slug: str

    :param language_slug: current GUI language slug
    :type language_slug: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    region = request.region
    event = get_object_or_404(region.events, id=event_id)

    event.archived = False
    event.save()

    # Restore related link objects
    for translation in event.translations.distinct("event__pk", "language__pk"):
        # The post_save signal will create link objects from the content
        translation.save(update_timestamp=False)

    logger.debug("%r restored by %r", event, request.user)
    messages.success(request, _("Event was successfully restored"))

    return redirect(
        "events",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@require_POST
@permission_required("cms.delete_event")
def delete(request, event_id, region_slug, language_slug):
    """
    Delete a single event

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param event_id: internal id of the event to be deleted
    :type event_id: int

    :param region_slug: slug of the region which the event belongs to
    :type region_slug: str

    :param language_slug: current GUI language slug
    :type language_slug: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    region = request.region
    event = get_object_or_404(region.events, id=event_id)

    logger.info("%r deleted by %r", event, request.user)

    if event.recurrence_rule:
        event.recurrence_rule.delete()
    event.delete()
    messages.success(request, _("Event was successfully deleted"))

    return redirect(
        "events",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@require_POST
@permission_required("cms.view_event")
# pylint: disable=unused-argument
def search_poi_ajax(request, region_slug):
    """
    AJAX endpoint for searching POIs

    :param request: Object representing the user call
    :type request: ~django.http.HttpRequest

    :param region_slug: The current regions slug
    :type region_slug: str

    :return: The rendered template response
    :rtype: ~django.template.response.TemplateResponse
    """
    data = json.loads(request.body.decode("utf-8"))

    poi_query = data.get("query_string")
    create_poi_option = data.get("create_poi_option")

    logger.debug('Ajax call: Live search for POIs with query "%r"', poi_query)

    region = get_object_or_404(Region, slug=data.get("region_slug"))

    # All latest revisions of a POI (one for each language)
    latest_public_poi_revisions = (
        POITranslation.objects.filter(poi=OuterRef("pk"), status=status.PUBLIC)
        .order_by("language__pk", "-version")
        .distinct("language")
        .values("id")
    )
    # All POIs which are not archived and have a latest public revision which contains the query
    poi_query_result = (
        region.pois.prefetch_related("translations")
        .filter(
            archived=False,
            translations__in=Subquery(latest_public_poi_revisions),
            translations__title__icontains=poi_query,
        )
        .distinct()
    )

    return render(
        request,
        "events/_poi_query_result.html",
        {
            "poi_query": poi_query,
            "poi_query_result": poi_query_result,
            "create_poi_option": create_poi_option,
            "region": region,
        },
    )



"""
We use :doc:`pytest <pytest:index>` to validate that the functionality of the cms works as expected.

In addition, we use the following plugins:

* :doc:`pytest-cov <pytest-cov:index>`: This plugin produces coverage reports.
* :doc:`pytest-django <pytest-django:index>`: Provide a few helpers for Django
* :doc:`pytest-xdist:index`: Enable distributing tests across multiple CPUs to speed up test execution

For more information, see :doc:`django:topics/testing/index` and :doc:`django:topics/testing/overview`.
"""



"""
APIv3 endpoint for feedback about the imprint
"""
import logging

from django.http import JsonResponse, Http404

from ....cms.models.feedback.imprint_page_feedback import ImprintPageFeedback
from ...decorators import json_response, feedback_handler

logger = logging.getLogger(__name__)


@feedback_handler
@json_response
def imprint_page_feedback(data, region, language, comment, rating, is_technical):
    """
    Store feedback about imprint in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :return: decorated function that saves feedback in database
    :rtype: ~collections.abc.Callable
    """
    return imprint_page_feedback_internal(
        data, region, language, comment, rating, is_technical
    )


# pylint: disable=unused-argument
def imprint_page_feedback_internal(
    data, region, language, comment, rating, is_technical
):
    """
    Store feedback about imprint in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :raises ~django.http.Http404: HTTP status 404 if the region has no imprint

    :return: JSON object according to APIv3 imprint feedback endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    if region.imprint:
        ImprintPageFeedback.objects.create(
            region=region,
            language=language,
            rating=rating,
            comment=comment,
            is_technical=is_technical,
        )
        return JsonResponse({"success": "Feedback successfully submitted"}, status=201)
    # If the corresponding imprint does not exist, return a 404 error
    logger.info(
        "The imprint for region %r in the language %s does not exist and no feedback can be given.",
        region,
        language,
    )
    raise Http404("The imprint does not exist in this region")



import logging
from django.shortcuts import get_object_or_404
from django.urls import resolve

from ...cms.models import Region

logger = logging.getLogger(__name__)


class RegionMiddleware:
    """
    Middleware class that adds the current region to the request variable
    """

    def __init__(self, get_response):
        """
        Initialize the middleware for the current view

        :param get_response: A callable to get the response for the current request
        :type get_response: ~collections.abc.Callable
        """
        self.get_response = get_response

    def __call__(self, request):
        """
        Call the middleware for the current request

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :return: The response after the region has been added to the request variable
        :rtype: ~django.http.HttpResponse
        """
        request.region = self.get_current_region(request)
        request.region_selection = self.get_region_selection(request)
        return self.get_response(request)

    @staticmethod
    def get_current_region(request):
        """
        This method returns the current region based on the current request.

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :raises ~django.http.Http404: When the current request has a ``region_slug`` parameter, but there is no region
                                      with that slug.

        :return: The current region of this request
        :rtype: ~integreat_cms.cms.models.regions.region.Region
        """
        # Resolve current url
        resolver_match = resolve(request.path)
        region_slug = resolver_match.kwargs.get("region_slug")
        if not region_slug:
            return None
        return get_object_or_404(Region, slug=region_slug)

    @staticmethod
    def get_region_selection(request):
        """
        This method returns the current region selection based on the current request.

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :return: The current region selection of this request
        :rtype: list [ ~integreat_cms.cms.models.regions.region.Region ]
        """
        # If user isn't authenticated, don't provide a region selection
        if not request.user.is_authenticated:
            return []
        # Firstly, attempt to provide the region selection based on the user's settings
        # (the regions should already have been prefetched by the custom user manager)
        region_selection = list(request.user.regions.all())[:10]
        # Exclude the current region from the available options
        if request.region:
            try:
                region_selection.remove(request.region)
            except ValueError:
                pass
        # If the user isn't a staff member or provided a valid selection, return those regions
        if (
            not request.user.is_superuser
            or not request.user.is_staff
            or len(region_selection) > 0
        ):
            return region_selection
        # As a last resort, query the 10 most recently updated regions from the database
        # (this is only relevant for staff members without preferred regions)
        region_selection = Region.objects.order_by("-last_updated")
        if request.region:
            region_selection = region_selection.exclude(pk=request.region.pk)
        return region_selection[:10]



"""
This package contains all views related to events
"""
from .event_form_view import EventFormView
from .event_list_view import EventListView
from .event_actions import (
    archive,
    restore,
    delete,
    search_poi_ajax,
    copy,
)



"""
This module includes functions related to the regions API endpoint.
"""
from django.http import JsonResponse

from ...cms.models import Region
from ...cms.constants import region_status
from ..decorators import json_response


def transform_region(region):
    """
    Function to create a JSON from a single region object, including information if region is live/active.

    :param region: The region object which should be converted
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :return: data necessary for API
    :rtype: dict
    """
    return {
        "id": region.id,
        "name": region.full_name,
        "path": region.slug,
        "live": region.status == region_status.ACTIVE,
        "prefix": region.prefix,
        "name_without_prefix": region.name,
        "plz": region.postal_code,
        "extras": region.offers.exists(),
        "events": region.events_enabled,
        "pois": region.locations_enabled,
        "push_notifications": region.push_notifications_enabled,
        "longitude": region.longitude,
        "latitude": region.latitude,
        "bounding_box": region.bounding_box.api_representation,
        "aliases": region.aliases,
        "tunews": region.tunews_enabled,
    }


def transform_region_by_status(region):
    """
    Function to create a JSON from a single "active" region object.

    :param region: The region object which should be converted
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :return: data necessary for API
    :rtype: dict
    """
    result = transform_region(region)
    # Remove status
    del result["live"]
    return result


@json_response
def regions(_):
    """
    List all regions that are not archived and transform result into JSON

    :return: JSON object according to APIv3 regions endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    result = list(
        map(transform_region, Region.objects.exclude(status=region_status.ARCHIVED))
    )
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays


@json_response
def liveregions(_):
    """
    List all regions that are not archived and transform result into JSON

    :return: JSON object according to APIv3 live regions endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    result = list(
        map(
            transform_region_by_status,
            Region.objects.filter(status=region_status.ACTIVE),
        )
    )
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays


@json_response
def hiddenregions(_):
    """
    List all regions that are hidden and transform result into JSON

    :return: JSON object according to APIv3 hidden regions endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    result = list(
        map(
            transform_region_by_status,
            Region.objects.filter(status=region_status.HIDDEN),
        )
    )
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays



"""
This module includes functions that extend the functionality of the Django Debug Toolbar to non HTML responses.
"""
import json
from django.http import HttpResponse


# pylint: disable=too-few-public-methods
class JsonDebugToolbarMiddleware:
    """
    The Django Debug Toolbar usually only works for views that return HTML.
    This middleware wraps any JSON response in HTML if the request
    has a 'debug' query parameter (e.g. http://localhost:8000/api/augsburg/de/pages?debug)
    """

    def __init__(self, get_response):
        """
        Initialize the middleware for the current view

        :param get_response: A callable to get the response for the current request
        :type get_response: ~collections.abc.Callable
        """
        self.get_response = get_response

    def __call__(self, request):
        """
        Call the middleware for the current request

        :param request: Django request
        :type request: ~django.http.HttpRequest

        :return: The modified response
        :rtype: ~django.http.HttpResponse
        """
        response = self.get_response(request)
        if "debug" in request.GET:
            if response["Content-Type"] == "application/json":
                content = json.dumps(
                    json.loads(response.content), sort_keys=True, indent=2
                )
                response = HttpResponse(
                    f"<!DOCTYPE html><html><body><pre>{content}</pre></body></html>"
                )

        return response



from urllib.parse import urlparse

from lxml.etree import LxmlError
from lxml.html import fromstring, tostring

from django import forms
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

from ..utils.slug_utils import generate_unique_slug_helper
from ..constants import status
from ..models import MediaFile
from .custom_model_form import CustomModelForm


class CustomContentModelForm(CustomModelForm):
    """
    Form for the content model forms for pages, events and POIs.
    """

    def __init__(self, **kwargs):
        r"""
        Initialize custom content model form

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """
        # Pop kwarg to make sure the super class does not get this param
        self.region = kwargs.pop("region", None)
        self.language = kwargs.pop("language", None)

        # Handle content edit lock
        self.changed_by_user = kwargs.pop("changed_by_user", None)
        self.locked_by_user = kwargs.pop("locked_by_user", None)

        # Instantiate CustomModelForm
        super().__init__(**kwargs)

        # The slug is not required because it will be auto-generated if left blank
        if "slug" in self.fields:
            self.fields["slug"].required = False

        if not self.locked_by_user:
            try:
                self.locked_by_user = self.instance.foreign_object.get_locking_user()
            except ObjectDoesNotExist:
                # If the foreign object does not exist yet, there ist also no lock so nothing must be done
                pass

    def clean(self):
        """
        This method extends the ``clean()``-method to verify that a user can modify this content model

        :return: The cleaned data (see :ref:`overriding-modelform-clean-method`)
        :rtype: dict
        """
        # Validate CustomModelForm
        cleaned_data = super().clean()

        force_update = cleaned_data.get("status") == status.AUTO_SAVE
        if (
            not force_update
            and self.changed_by_user
            and self.locked_by_user
            and self.changed_by_user != self.locked_by_user
        ):
            self.add_error(
                None,
                forms.ValidationError(
                    _(
                        "Could not update because this content because it is already being edited by another user"
                    ),
                    code="invalid",
                ),
            )

        return cleaned_data

    def clean_content(self):
        """
        Validate the content field (see :ref:`overriding-modelform-clean-method`) and applies changes
        to ``<img>``- and ``<a>``-Tags to match the guidelines.

        :raises ~django.core.exceptions.ValidationError: When a heading 1 (``<h1>``) is used in the text content

        :return: The valid content
        :rtype: str
        """
        try:
            content = fromstring(self.cleaned_data["content"])
        except LxmlError:
            # The content is not guaranteed to be valid html, for example it may be empty
            return self.cleaned_data["content"]

        # Convert heading 1 to heading 2
        for heading in content.iter("h1"):
            heading.tag = "h2"
            self.logger.debug(
                "Replaced heading 1 with heading 2: %r", tostring(heading)
            )

        # Convert pre and code tags to p tags
        for monospaced in content.iter("pre", "code"):
            tag_type = monospaced.tag
            monospaced.tag = "p"
            self.logger.debug(
                "Replaced %r tag with p tag: %r", tag_type, tostring(monospaced)
            )

        # Remove external links
        for link in content.iter("a"):
            link.attrib.pop("target", None)
            self.logger.debug("Removed target attribute from link: %r", tostring(link))

        # Scan for media files in content and replace alt texts
        for image in content.iter("img"):
            self.logger.debug(
                "Image tag found in content (src: %s)", image.attrib["src"]
            )
            # Remove host
            relative_url = urlparse(image.attrib["src"]).path
            # Remove media url prefix if exists
            if relative_url.startswith(settings.MEDIA_URL):
                relative_url = relative_url[len(settings.MEDIA_URL) :]
            # Check whether media file exists in database
            media_file = MediaFile.objects.filter(
                Q(file=relative_url) | Q(thumbnail=relative_url)
            ).first()
            # Replace alternative text
            if media_file and media_file.alt_text:
                self.logger.debug("Image alt text replaced: %r", media_file.alt_text)
                image.attrib["alt"] = media_file.alt_text

        return tostring(content, with_tail=False).decode("utf-8")

    def clean_slug(self):
        """
        Validate the slug field (see :ref:`overriding-modelform-clean-method`)

        :return: A unique slug based on the input value
        :rtype: str
        """
        unique_slug = generate_unique_slug_helper(
            self, self._meta.model.foreign_field()
        )
        self.data = self.data.copy()
        self.data["slug"] = unique_slug
        return unique_slug

    # pylint: disable=arguments-differ
    def save(self, commit=True):
        """
        This method extends the default ``save()``-method of the base :class:`~django.forms.ModelForm` to set attributes
        which are not directly determined by input fields.

        :param commit: Whether or not the changes should be written to the database
        :type commit: bool

        :return: The saved page translation object
        :rtype: ~integreat_cms.cms.models.pages.page_translation.PageTranslation
        """

        # Delete now outdated link objects
        self.instance.links.all().delete()

        # If none of the text content fields changed, treat as minor edit (even if checkbox isn't clicked)
        if {"title", "short_description", "content"}.isdisjoint(self.changed_data):
            self.logger.debug("Set 'minor_edit=True' since the content did not change")
            self.instance.minor_edit = True

        # Save new version
        self.instance.version += 1
        self.instance.pk = None

        # Save CustomModelForm
        return super().save(commit=commit)

    def add_success_message(self, request):
        """
        This adds a success message for a translation form.
        Requires the attributes "title", "status" and "foreign_object" on the form instance.

        :param request: The current request submitting the translation form
        :type request: ~django.http.HttpRequest
        """
        model_name = type(self.instance.foreign_object)._meta.verbose_name.title()
        if not self.instance.status == status.PUBLIC:
            messages.success(
                request,
                _('{} "{}" was successfully saved as draft').format(
                    model_name, self.instance.title
                ),
            )
        elif "status" not in self.changed_data:
            messages.success(
                request,
                _('{} "{}" was successfully updated').format(
                    model_name, self.instance.title
                ),
            )
        else:
            messages.success(
                request,
                _('{} "{}" was successfully published').format(
                    model_name, self.instance.title
                ),
            )



from django.core.validators import MinLengthValidator
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _

from cacheops import invalidate_obj

from ..abstract_base_model import AbstractBaseModel
from ...constants import countries, text_directions
from ...utils.translation_utils import ugettext_many_lazy as __


class Language(AbstractBaseModel):
    """
    Data model representing a content language.
    """

    slug = models.SlugField(
        max_length=8,
        unique=True,
        validators=[MinLengthValidator(2)],
        verbose_name=_("Language Slug"),
        help_text=_(
            "Unique string identifier used in URLs without spaces and special characters."
        ),
    )
    #: The recommended minimum buffer for `bcp47 <https://tools.ietf.org/html/bcp47>`__ is 35.
    #: It's unlikely that we have language slugs longer than 8 characters though.
    #: See `RFC 5646 Section 4.4.1. <https://tools.ietf.org/html/bcp47#section-4.4.1>`__
    #: Registered tags: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry
    bcp47_tag = models.SlugField(
        max_length=35,
        unique=True,
        validators=[MinLengthValidator(2)],
        verbose_name=_("BCP47 Tag"),
        help_text=__(
            _("Language identifier without spaces and special characters."),
            _(
                "This field usually contains a combination of subtags from the IANA Subtag Registry."
            ),
        ),
    )
    native_name = models.CharField(
        max_length=250,
        blank=False,
        verbose_name=_("native name"),
        help_text=_("The name of the language in this language."),
    )
    english_name = models.CharField(
        max_length=250,
        blank=False,
        verbose_name=_("name in English"),
        help_text=_("The name of the language in English."),
    )
    #: Manage choices in :mod:`~integreat_cms.cms.constants.text_directions`
    text_direction = models.CharField(
        default=text_directions.LEFT_TO_RIGHT,
        choices=text_directions.CHOICES,
        max_length=13,
        verbose_name=_("text direction"),
    )
    #: Manage choices in :mod:`~integreat_cms.cms.constants.countries`
    primary_country_code = models.CharField(
        choices=countries.CHOICES,
        max_length=2,
        verbose_name=_("primary country flag"),
        help_text=__(
            _("The country with which this language is mainly associated."),
            _("This flag is used to represent the language graphically."),
        ),
    )
    #: Manage choices in :mod:`~integreat_cms.cms.constants.countries`
    secondary_country_code = models.CharField(
        choices=countries.CHOICES,
        blank=True,
        max_length=2,
        verbose_name=_("secondary country flag"),
        help_text=__(
            _("Another country with which this language is also associated."),
            _("This flag is used in the language switcher."),
        ),
    )
    created_date = models.DateTimeField(
        default=timezone.now,
        verbose_name=_("creation date"),
    )
    last_updated = models.DateTimeField(
        auto_now=True,
        verbose_name=_("modification date"),
    )
    table_of_contents = models.CharField(
        max_length=250,
        blank=False,
        verbose_name=_('"Table of contents" in this language'),
        help_text=__(
            _('The native name for "Table of contents" in this language.'),
            _("This is used in exported PDFs."),
        ),
    )
    message_content_not_available = models.CharField(
        max_length=250,
        blank=False,
        default="This page does not exist in the selected language. It is however available in these languages:",
        verbose_name=_(
            '"This page does not exist in the selected language. It is however available in these languages:" in this language'
        ),
        help_text=_(
            "This is shown to the user when a page is not available in their language."
        ),
    )
    message_partial_live_content_not_available = models.CharField(
        max_length=250,
        blank=False,
        default="Part of the page does not exist in the selected language. It is however available in these languages:",
        verbose_name=_(
            '"Part of the page does not exist in the selected language. It is however available in these languages:" in this language'
        ),
        help_text=_(
            "This is shown to the user when the mirrored part of a page is not available in their language."
        ),
    )

    @cached_property
    def translated_name(self):
        """
        Returns the name of the language in the current backend language

        :return: The translated name of the language
        :rtype: str
        """
        return ugettext(self.english_name)

    def save(self, *args, **kwargs):
        r"""
        This overwrites the default Django :meth:`~django.db.models.Model.save` method,
        to invalidate the cache of the related objects.

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied kwargs
        :type \**kwargs: dict
        """
        super().save(*args, **kwargs)
        # Invalidate related objects
        for obj in self.language_tree_nodes.all():
            invalidate_obj(obj)
        for obj in self.page_translations.all():
            invalidate_obj(obj)
        for obj in self.event_translations.all():
            invalidate_obj(obj)
        for obj in self.poi_translations.all():
            invalidate_obj(obj)
        for obj in self.push_notification_translations.all():
            invalidate_obj(obj)

    def __str__(self):
        """
        This overwrites the default Django :meth:`~django.db.models.Model.__str__` method which would return ``Language object (id)``.
        It is used in the Django admin backend and as label for ModelChoiceFields.

        :return: A readable string representation of the language
        :rtype: str
        """
        return self.translated_name

    def get_repr(self):
        """
        This overwrites the default Django ``__repr__()`` method which would return ``<Language: Language object (id)>``.
        It is used for logging.

        :return: The canonical string representation of the language
        :rtype: str
        """
        return (
            f"<Language (id: {self.id}, slug: {self.slug}, name: {self.english_name})>"
        )

    class Meta:
        #: The verbose name of the model
        verbose_name = _("language")
        #: The plural verbose name of the model
        verbose_name_plural = _("languages")
        #: The default permissions for this model
        default_permissions = ("change", "delete", "view")
        #: The fields which are used to sort the returned objects of a QuerySet
        ordering = ["bcp47_tag"]



from django.conf import settings

from linkcheck import Linklist

from .models import Page, PageTranslation, EventTranslation, POITranslation


class ContentTranslationLinklist(Linklist):
    """
    Base class for content translation link lists
    """

    html_fields = ["content"]

    @classmethod
    def filter_callable(cls, objects):
        """
        Get only latest versions for non-archived events/locations

        :param objects: Objects to be filtered
        :type objects: ~django.db.models.query.QuerySet

        :return: Objects that passed the filter
        :rtype: ~django.db.models.query.QuerySet
        """
        return objects.filter(
            **{f"{objects.model.foreign_field()}__archived": False}
        ).distinct(f"{objects.model.foreign_field()}__pk", "language__pk")


class PageTranslationLinklist(Linklist):
    """
    Class for selecting the PageTranslation model for link checks
    """

    model = PageTranslation
    html_fields = ["content"]

    @classmethod
    def filter_callable(cls, objects):
        """
        Get only latest versions for non-archived pages

        :param objects: Objects to be filtered
        :type objects: ~django.db.models.query.QuerySet

        :return: Objects that passed the filter
        :rtype: ~django.db.models.query.QuerySet
        """
        # Since filtering the archived pages causes a lot of overhead, we only want this during the findlinks management command.
        if settings.LINKCHECK_EXCLUDE_ARCHIVED_PAGES:
            pages = Page.objects.all().cache_tree(archived=False)
            objects = objects.filter(page__in=pages)
        return objects.distinct("page__pk", "language__pk")


class EventTranslationLinklist(ContentTranslationLinklist):
    """
    Class for selecting the EventTranslation model for link checks
    """

    model = EventTranslation


class POITranslationLinklist(ContentTranslationLinklist):
    """
    Class for selecting the POITranslation model for link checks
    """

    model = POITranslation


linklists = {
    "PageTranslations": PageTranslationLinklist,
    "EventTranslations": EventTranslationLinklist,
    "POITranslations": POITranslationLinklist,
}



import logging

from urllib.parse import urlparse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import override

from geopy.geocoders import Nominatim
from geopy.exc import GeopyError

from integreat_cms import __version__

from ..cms.constants import administrative_division as ad
from .utils import BoundingBox

logger = logging.getLogger(__name__)


class NominatimApiClient:
    """
    Client to interact with the Nominatim API.
    For documentation about the underlying library, see :doc:`GeoPy <geopy:index>`.
    """

    def __init__(self):
        """
        Initialize the Nominatim client

        :raises ~django.core.exceptions.ImproperlyConfigured: When the Nominatim API is disabled
        """
        if not settings.NOMINATIM_API_ENABLED:
            raise ImproperlyConfigured("Nominatim API is disabled")
        try:
            nominatim_url = urlparse(settings.NOMINATIM_API_URL)
            self.geolocator = Nominatim(
                domain=nominatim_url.netloc + nominatim_url.path,
                scheme=nominatim_url.scheme,
                user_agent=f"integreat-cms/{__version__} ({settings.HOSTNAME})",
            )
        except GeopyError as e:
            logger.exception(e)
            logger.error("Nominatim API client could not be initialized")

    def search(self, query_str=None, exactly_one=True, **query_dict):
        r"""
        Search for a given query, either by string or by dict.
        ``query_str`` and ``query_dict`` are mutually exclusive.

        :raises RuntimeError: When the ``query_str`` and ``query_dict`` parameters are passed at the same time

        :param query_str: The query string
        :type query_str: str

        :param exactly_one: Whether only one result should be returned
        :type exactly_one: bool

        :param \**query_dict: The query as dictionary
        :type \**query_dict: dict

        :return: The location that matches the given criteria
        :rtype: geopy.location.Location
        """
        if query_str and query_dict:
            raise RuntimeError(
                "You can either specify query_str or pass additional keyword arguments, not both."
            )
        query = query_str or query_dict
        try:
            result = self.geolocator.geocode(
                query,
                exactly_one=exactly_one,
            )
            if result:
                logger.debug("Nominatim API search result: %r", result.raw)
            else:
                logger.debug("Nominatim API did not return a match")
            return result
        except GeopyError as e:
            logger.exception(e)
            logger.error("Nominatim API call failed")
            return None

    def check_availability(self):
        """
        Check if Nominatim API is available
        """
        try:
            assert self.search(query_str="Deutschland")
            logger.info(
                "Nominatim API is available at: %r",
                settings.NOMINATIM_API_URL,
            )
        except (GeopyError, AssertionError) as e:
            logger.exception(e)
            logger.error(
                "Nominatim API unavailable. You won't be able to "
                "automatically import location coordinates."
            )

    def get_coordinates(self, street, postalcode, city):
        """
        Get coordinates for given address

        :param street: The requested street
        :type street: str

        :param postalcode: The requested postal code
        :type postalcode: str

        :param city: The requested city
        :type city: str

        :return: The coordinates of the requested address
        :rtype: tuple ( float, float )
        """
        result = self.search(street=street, postalcode=postalcode, city=city)
        if result:
            return result.latitude, result.longitude
        return None, None

    def get_bounding_box(self, administrative_division, name, aliases=None):
        """
        Get the bounding box for a given region

        :param administrative_division: The administrative division of the requested region
        :type administrative_division: str

        :param name: The name of the requested region
        :type name: str

        :param aliases: A dictionary of region aliases
        :type aliases: dict

        :return: The bounding box
        :rtype: ~integreat_cms.nominatim_api.utils.BoundingBox
        """
        if administrative_division in [
            ad.CITY,
            ad.MUNICIPALITY,
            ad.CITY_STATE,
            ad.URBAN_DISTRICT,
            ad.COLLECTIVE_MUNICIPALITY,
        ]:
            return self.get_city_bounding_box(name)
        if administrative_division == ad.CITY_AND_DISTRICT:
            return self.get_city_and_district_bounding_box(name)
        if administrative_division in [ad.DISTRICT, ad.RURAL_DISTRICT]:
            return self.get_district_bounding_box(administrative_division, name)
        if administrative_division == ad.REGION:
            return self.get_region_bounding_box(name, aliases)
        return None

    def get_city_bounding_box(self, name):
        """
        Get the bounding box for a given city

        :param name: The name of the requested region
        :type name: str

        :return: The bounding box
        :rtype: ~integreat_cms.nominatim_api.utils.BoundingBox
        """
        # For cities and municipalities, we can just use the "city" parameter
        return BoundingBox.from_result(self.search(city=name))

    def get_city_and_district_bounding_box(self, name):
        """
        Get the bounding box for a given city and district

        :param name: The name of the requested region
        :type name: str

        :return: The bounding box
        :rtype: ~integreat_cms.nominatim_api.utils.BoundingBox
        """
        # Get bounding box of city
        city_box = BoundingBox.from_result(self.search(city=name))
        # Get bounding box of district
        district_box = BoundingBox.from_result(
            self.search(query_str=f"Landkreis {name}")
        )
        # Merge both results
        return BoundingBox.merge(city_box, district_box)

    def get_district_bounding_box(self, administrative_division, name):
        """
        Get the bounding box for a given district

        :param administrative_division: The administrative division of the requested region
        :type administrative_division: str

        :param name: The name of the requested region
        :type name: str

        :return: The bounding box
        :rtype: ~integreat_cms.nominatim_api.utils.BoundingBox
        """
        # Get translated name of the administrative division
        with override(settings.LANGUAGE_CODE):
            adm_div = dict(ad.CHOICES)[administrative_division]
        # For districts, we have to use string queries
        return BoundingBox.from_result(self.search(query_str=f"{adm_div} {name}"))

    def get_region_bounding_box(self, name, aliases=None):
        """
        Get the bounding box for a given region and all its aliases

        :param name: The name of the requested region
        :type name: str

        :param aliases: A dictionary of region aliases
        :type aliases: dict

        :return: The bounding box
        :rtype: ~integreat_cms.nominatim_api.utils.BoundingBox
        """
        if not aliases:
            aliases = {}
        # Get bounding box of city
        bounding_boxes = [BoundingBox.from_result(self.search(city=name))]
        # Get bounding boxes of all aliases
        for alias in aliases.keys():
            bounding_boxes.append(BoundingBox.from_result(self.search(city=alias)))
        return BoundingBox.merge(*bounding_boxes)



"""
Override the app config of the linkcheck app
"""
from linkcheck.apps import LinkcheckConfig


class ModifiedLinkcheckConfig(LinkcheckConfig):
    """
    Use the legacy field for automatic primary keys
    """

    default_auto_field = "django.db.models.AutoField"



"""
This module contains all views related to multi-factor authentication
"""
import logging

from django.contrib import messages
from django.shortcuts import render, redirect
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ....decorators import modify_mfa_authenticated

logger = logging.getLogger(__name__)


@method_decorator(modify_mfa_authenticated, name="dispatch")
class DeleteUserMfaKeyView(TemplateView):
    """
    View to delete a multi-factor-authentication key
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "settings/mfa/delete.html"

    def get(self, request, *args, **kwargs):
        r"""
        Render mfa-deletion view

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        key = request.user.mfa_keys.get(id=kwargs["key_id"])

        if request.user.mfa_keys.count() == 1:
            messages.warning(
                request,
                _(
                    "This is your last key, once removed you will be able to log in without a second factor."
                ),
            )
        else:
            messages.warning(
                request,
                " ".join(
                    [
                        _(
                            "Once you remove the key you will need to use one of the other available keys to log into your account."
                        ),
                        _(
                            "Please make sure that you have at least one extra key available to log in before removing this key."
                        ),
                    ]
                ),
            )

        return render(
            request,
            self.template_name,
            {"key": key},
        )

    def post(self, request, **kwargs):
        r"""
        Delete a multi-factor-authentication key

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: A redirection to the user settings
        :rtype: ~django.http.HttpResponseRedirect
        """

        key = request.user.mfa_keys.get(id=kwargs["key_id"])
        messages.success(
            request,
            _('The 2-factor authentication key "{}" was successfully deleted').format(
                key.name
            ),
        )
        key.delete()
        kwargs = (
            {"region_slug": self.request.region.slug} if self.request.region else {}
        )
        return redirect("user_settings", **kwargs)



import logging

from django.contrib import messages
from django.contrib.auth import views as auth_views
from django.utils.translation import ugettext as _

logger = logging.getLogger(__name__)


class LogoutView(auth_views.LogoutView):
    """
    View to log off a user
    """

    def dispatch(self, request, *args, **kwargs):
        r"""
        This function sends message, if logout was successful

        :param request: The current request
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: passes form to dispatch function
        :rtype: ~django.http.HttpResponse
        """
        messages.info(request, _("You have been successfully logged off."))
        return super().dispatch(request, *args, **kwargs)



import logging

from email.mime.image import MIMEImage

from django.conf import settings
from django.contrib.staticfiles import finders
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils.text import capfirst

logger = logging.getLogger(__name__)


def send_mail(
    subject, email_template_name, html_email_template_name, context, recipient
):
    """
    Sends welcome email to user with new account.

    Making use of  :class:`~django.core.mail.EmailMultiAlternatives` and
    :class:`~email.mime.image.MIMEImage` to send a multipart email with plain text and html
    version and also attach the branding logo.

    :param subject: The subject of the email
    :type subject: str

    :param email_template_name: The template to be used to render the text email
    :type email_template_name: str

    :param html_email_template_name: The template to be used to render the HTML email
    :type html_email_template_name: str

    :param subject: The subject of the email
    :type subject: str

    :param context: The template context variables
    :type context: dict

    :param recipient: The email address of the recipient
    :type recipient: str
    """
    context.update(
        {
            "base_url": settings.BASE_URL,
            "COMPANY": settings.COMPANY,
            "BRANDING": capfirst(settings.BRANDING),
        }
    )
    # Assemble message body
    body = render_to_string(
        email_template_name,
        context,
    )
    # Initialize message
    email = EmailMultiAlternatives(subject=subject, body=body, to=[recipient])
    email.mixed_subtype = "related"
    # Attach HTML email body
    html_message = render_to_string(html_email_template_name, context)
    email.attach_alternative(html_message, "text/html")
    # Attach logo
    image_path = finders.find(f"logos/{settings.BRANDING}/{settings.BRANDING}-logo.png")
    if image_path:
        with open(image_path, mode="rb") as f:
            image = MIMEImage(f.read())
            email.attach(image)
            image.add_header("Content-ID", "<logo>")
    else:
        logger.debug(
            "Logo not found at %s, will proceed without attaching.",
            finders.searched_locations,
        )
    # Send email
    email.send()



import logging
import time

from copy import deepcopy
from functools import partial
from urllib.parse import urlencode

from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect, reverse
from django.utils.translation import gettext as _
from django.views.generic import ListView

from lxml.html import rewrite_links

from linkcheck import update_lock
from linkcheck.models import Link, Url
from cacheops import invalidate_model

from ...utils.linkcheck_utils import filter_urls, get_urls
from ...forms.linkcheck.edit_url_form import EditUrlForm

logger = logging.getLogger(__name__)


class LinkcheckListView(ListView):
    """
    View for retrieving a list of urls grouped by their state
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "linkcheck/links_by_filter.html"
    #: Designates the name of the variable to use in the context
    #: (see :class:`~django.views.generic.list.MultipleObjectMixin`)
    context_object_name = "filtered_urls"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {"current_menu_item": "linkcheck"}
    #: The link edit form
    form = None

    def get_paginate_by(self, queryset):
        """
        Get the number of items to paginate by, or ``None`` for no pagination.

        :param queryset: The QuerySet of the filtered urls
        :type queryset: ~django.db.models.query.QuerySet

        :return: The pagination number
        :rtype: int
        """
        return int(self.request.GET.get("size", settings.PER_PAGE))

    def get_pagination_params(self):
        """
        Get the urlencoded pagination GET parameters

        :return: The URL params
        :rtype: str
        """
        params = {
            k: v for k, v in self.request.GET.items() if k in (self.page_kwarg, "size")
        }
        print(params)
        if params:
            return f"?{urlencode(params)}"
        return ""

    def get_context_data(self, **kwargs):
        r"""
        Extend context by edit url form

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The context dictionary
        :rtype: dict
        """
        context = super().get_context_data(**kwargs)
        edit_url_id = self.kwargs.get("url_id")
        if edit_url_id and not self.form:
            url = get_object_or_404(Url, id=edit_url_id)
            self.form = EditUrlForm(initial={"url": url})
        context["edit_url_form"] = self.form
        context["pagination_params"] = self.get_pagination_params()
        return context

    def get_queryset(self):
        """
        Selects all urls for the current region
        Finally annotates queryset by the content_type title

        :return: The QuerySet of the filtered urls
        :rtype: ~django.db.models.query.QuerySet
        """
        urls, count_dict = filter_urls(
            self.kwargs.get("region_slug"), self.kwargs.get("url_filter")
        )
        self.extra_context.update(count_dict)
        return urls

    def get(self, request, *args, **kwargs):
        r"""
        Return the get view and fall back to the last page if the requested page does not exist

        :param request: Object representing the user call
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: When the view returns a 404 and already requests the last page

        :return: The rendered linkcheck template or a redirect to the last page
        :rtype: ~django.template.response.TemplateResponse or ~django.http.HttpResponseRedirect
        """
        try:
            return super().get(request, *args, **kwargs)
        except Http404 as e:
            # If already the last page was requested, raise the error
            if request.GET.get("page") == "last":
                raise (e)
            # If the page does not exist, use the last page as fallback
            logger.debug("Redirecting to last page because response was 404")
            params = {"page": "last"}
            size = request.GET.get("size")
            if size:
                params["size"] = size
            return redirect(f"{request.path}?{urlencode(params)}")

    # pylint: disable-msg=too-many-locals,too-many-branches
    def post(self, request, *args, **kwargs):
        r"""
        Applies selected action for selected urls

        :param request: The current request
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: HTTP status 404 if the edited URL does not exist

        :return: Redirect to current linkcheck tab
        :rtype: ~django.http.HttpResponseRedirect
        """
        edit_url_id = kwargs.pop("url_id", None)
        if edit_url_id:
            try:
                old_url = get_urls(
                    region_slug=request.region.slug, url_ids=[edit_url_id]
                )[0]
            except IndexError as e:
                raise Http404("This URL does not exist") from e
            self.form = EditUrlForm(data=request.POST)
            if self.form.is_valid():
                new_url = self.form.cleaned_data["url"]
                # Get all current translations with the same url
                translations = {link.content_object for link in old_url.region_links}
                # Replace the old urls with the new urls in the content
                for translation in translations:
                    new_translation = deepcopy(translation)
                    # Replace link in translation
                    logger.debug("Replacing links of %r", new_translation)
                    new_translation.content = rewrite_links(
                        new_translation.content,
                        partial(self.replace_link, old_url.url, new_url),
                    )
                    # Save translation with replaced content as new minor version
                    new_translation.id = None
                    new_translation.version += 1
                    new_translation.minor_edit = True
                    new_translation.save()
                messages.success(request, _("URL was successfully replaced"))
                # Add short delay to allow post_save signals to finish (to keep existing URL objects when deleting the old links)
                time.sleep(0.5)
                # Acquire linkcheck lock to avoid race conditions between post_save signal and links.delete()
                with update_lock:
                    for translation in translations:
                        # Delete now outdated link objects
                        translation.links.all().delete()
                # Add short delay to allow rechecking to be finished when page reloads
                time.sleep(0.5)
            else:
                # Show error messages
                for field in self.form:
                    for error in field.errors:
                        messages.error(request, _(field.label) + ": " + _(error))
                # If the form is invalid, render the invalid form
                return super().get(request, *args, **kwargs)

        action = request.POST.get("action")
        selected_urls = get_urls(
            region_slug=request.region.slug,
            url_ids=request.POST.getlist("selected_ids[]"),
            prefetch_content_objects=False,
        )

        if action == "ignore":
            for url in selected_urls:
                Link.objects.filter(
                    id__in=[link.id for link in url.region_links]
                ).update(ignore=True)
            messages.success(request, _("Links were successfully ignored"))
        elif action == "unignore":
            for url in selected_urls:
                Link.objects.filter(
                    id__in=[link.id for link in url.region_links]
                ).update(ignore=False)
            messages.success(request, _("Links were successfully unignored"))
        elif action == "recheck":
            for url in selected_urls:
                url.check_url(external_recheck_interval=0)
            messages.success(request, _("Links were successfully checked"))
            # Add short delay to allow rechecking to be finished when page reloads
            time.sleep(1)
        invalidate_model(Link)
        invalidate_model(Url)
        linkcheck_url = reverse("linkcheck", kwargs=kwargs)
        # Keep pagination settings
        return redirect(f"{linkcheck_url}{self.get_pagination_params()}")

    @staticmethod
    def replace_link(old_url, new_url, link):
        """
        Replace the URL of a link

        :param old_url: The old URL to be replaced
        :type old_url: str

        :param new_url: The new URL
        :type new_url: str

        :param link: The input link
        :type link: str

        :return: The replaced link
        :rtype: str
        """
        if link == old_url:
            logger.debug("Replacing %r with %r", old_url, new_url)
            return new_url
        return link



"""
This module contains shared fixtures for pytest
"""
import pytest

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.test.client import Client, AsyncClient


from integreat_cms.cms.constants.roles import (
    MANAGEMENT,
    EDITOR,
    AUTHOR,
    EVENT_MANAGER,
    SERVICE_TEAM,
    CMS_TEAM,
    APP_TEAM,
    MARKETING_TEAM,
)

#: A role identifier for superusers
ROOT = "ROOT"
#: A role identifier for anonymous users
ANONYMOUS = "ANONYMOUS"

#: All roles of region users
REGION_ROLES = [MANAGEMENT, EDITOR, AUTHOR, EVENT_MANAGER]
#: All roles of staff users
STAFF_ROLES = [ROOT, SERVICE_TEAM, CMS_TEAM, APP_TEAM, MARKETING_TEAM]
#: All roles of staff users that don't just have read-only permissions
PRIV_STAFF_ROLES = [ROOT, APP_TEAM, SERVICE_TEAM, CMS_TEAM]
#: All roles of staff users that don't just have read-only permissions
HIGH_PRIV_STAFF_ROLES = [ROOT, SERVICE_TEAM, CMS_TEAM]
#: All region and staff roles
ROLES = REGION_ROLES + STAFF_ROLES
#: All region and staff roles and anonymous users
ALL_ROLES = ROLES + [ANONYMOUS]

#: Enable the aiohttp pytest plugin to make use of the test server
pytest_plugins = "aiohttp.pytest_plugin"

# pylint: disable=unused-argument
@pytest.fixture(scope="session")
def load_test_data(django_db_setup, django_db_blocker):
    """
    Load the test data initially for all test cases

    :param django_db_setup: The fixture providing the database availability
    :type django_db_setup: :fixture:`django_db_setup`

    :param django_db_blocker: The fixture providing the database blocker
    :type django_db_blocker: :fixture:`django_db_blocker`
    """
    with django_db_blocker.unblock():
        call_command("loaddata", "integreat_cms/cms/fixtures/test_data.json")


# pylint: disable=unused-argument,redefined-outer-name
@pytest.fixture(scope="session", params=ALL_ROLES)
def login_role_user(request, load_test_data, django_db_blocker):
    """
    Get the test user of the current role and force a login. Gets executed only once per user.

    :param request: The request object providing the parametrized role variable through ``request.param``
    :type request: pytest.FixtureRequest

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: NoneType

    :param django_db_blocker: The fixture providing the database blocker
    :type django_db_blocker: :fixture:`django_db_blocker`

    :return: The http client and the current role
    :rtype: tuple
    """
    client = Client()
    # Only log in user if the role is not anonymous
    if request.param != ANONYMOUS:
        with django_db_blocker.unblock():
            user = get_user_model().objects.get(username=request.param.lower())
            client.force_login(user)
    return client, request.param


# pylint: disable=unused-argument,redefined-outer-name
@pytest.fixture(scope="session", params=ALL_ROLES)
def login_role_user_async(request, load_test_data, django_db_blocker):
    """
    Get the test user of the current role and force a login. Gets executed only once per user.
    Identical to :meth:`~tests.conftest.login_role_user` with the difference that it returns
    an :class:`django.test.client.AsyncClient` instead of :class:`django.test.client.Client`.

    :param request: The request object providing the parametrized role variable through ``request.param``
    :type request: pytest.FixtureRequest

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: NoneType

    :param django_db_blocker: The fixture providing the database blocker
    :type django_db_blocker: :fixture:`django_db_blocker`

    :return: The http client and the current role
    :rtype: tuple
    """
    async_client = AsyncClient()
    # Only log in user if the role is not anonymous
    if request.param != ANONYMOUS:
        with django_db_blocker.unblock():
            user = get_user_model().objects.get(username=request.param.lower())
            async_client.force_login(user)
    return async_client, request.param



import pytest

from asgiref.sync import sync_to_async
from aiohttp import web

from django.urls import reverse

from integreat_cms.cms.models import Page, Region

from ..conftest import (
    ANONYMOUS,
    PRIV_STAFF_ROLES,
    MANAGEMENT,
    EDITOR,
    AUTHOR,
)


async def fake_summ_ai_server(request):
    """
    Create fake responses which simulate the SUMM.AI API server

    :param request: The request
    :type request: aiohttp.web.Request

    :return: The response
    :rtype: aiohttp.web.Response
    """
    return web.json_response(
        {
            "translated_text": "Hier ist Ihre Leichte Sprache Übersetzung",
            "jobid": "9999",
        }
    )


@sync_to_async
def enable_summ_api(region_slug):
    """
    Enable SUMM.AI in the test region

    :param region_slug: The slug of the region in which we want to enable SUMM.AI
    :type region_slug: str
    """
    # Enable SUMM.AI in the test region
    region = Region.objects.get(slug=region_slug)
    region.summ_ai_enabled = True
    region.save()


@sync_to_async
def get_changed_pages(settings, ids):
    """
    Load the changed pages with the specified ids from the database

    :param settings: The fixture providing the django settings
    :type settings: :fixture:`settings`

    :param ids: A list containing the requested pages including their translations in German and Easy German
    :type ids: list [ int ]

    :return: The changed pages
    :rtype: list [ dict ]
    """
    # Enable SUMM.AI in the test region
    return [
        {
            "page": page,
            **{
                slug: page.get_translation(slug)
                for slug in [
                    settings.SUMM_AI_GERMAN_LANGUAGE_SLUG,
                    settings.SUMM_AI_EASY_GERMAN_LANGUAGE_SLUG,
                ]
            },
        }
        for page in Page.objects.filter(id__in=ids)
    ]


@pytest.mark.django_db
# pylint: disable=too-many-locals
async def test_auto_translate_easy_german(
    login_role_user_async, settings, aiohttp_raw_server
):
    """
    This test checks whether the SUMM.AI API client works as expected

    :param login_role_user_async: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user_async: tuple

    :param settings: The fixture providing the django settings
    :type settings: :fixture:`settings`

    :param aiohttp_raw_server: The fixture providing the dummy aiohttp server used for faking the SUMM.AI API server
    :type aiohttp_raw_server: :data:`aiohttp:pytest_aiohttp.aiohttp_raw_server`
    """
    # The region we want to use for testing
    region_slug = "augsburg"
    # Enable SUMM.AI in the test region
    await enable_summ_api(region_slug)
    # Setup a dummy server to fake responses from the SUMM.AI API server (an instance of aiohttp.test_utils.RawTestServer)
    fake_server = await aiohttp_raw_server(fake_summ_ai_server)
    # Redirect call to the SUMM.AI API to the fake server
    settings.SUMM_AI_API_URL = (
        f"{fake_server.scheme}://{fake_server.host}:{fake_server.port}"
    )
    # Test for english messages
    settings.LANGUAGE_CODE = "en"
    # Log the user in
    client, role = login_role_user_async
    # Translate the pages
    selected_ids = [1, 2, 3]
    translate_easy_german = reverse(
        "auto_translate_easy_german_pages",
        kwargs={"region_slug": region_slug, "language_slug": "de"},
    )
    response = await client.post(
        translate_easy_german, data={"selected_ids[]": selected_ids}
    )
    print(response.headers)
    if role in PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]:
        # If the role should be allowed to access the view, we expect a successful result
        assert response.status_code == 302
        page_tree = reverse(
            "pages", kwargs={"region_slug": region_slug, "language_slug": "de"}
        )
        assert response.headers.get("Location") == page_tree
        response = await client.get(page_tree)
        print(response.headers)
        # Get the page objects including their translations from the database
        changed_pages = await get_changed_pages(settings, selected_ids)
        for page in changed_pages:
            # Check that the success message are present
            assert (
                f'Page "{page[settings.SUMM_AI_GERMAN_LANGUAGE_SLUG]}" has been successfully translated into Easy German.'
                in response.content.decode("utf-8")
            )
            # Check that the page translation exists and really has the correct content
            assert (
                "Hier ist Ihre Leichte Sprache &#220;bersetzung"
                in page[settings.SUMM_AI_EASY_GERMAN_LANGUAGE_SLUG].content
            )
    elif role == ANONYMOUS:
        # For anonymous users, we want to redirect to the login form instead of showing an error
        assert response.status_code == 302
        assert (
            response.headers.get("location")
            == f"{settings.LOGIN_URL}?next={translate_easy_german}"
        )
    else:
        # For logged in users, we want to show an error if they get a permission denied
        assert response.status_code == 403


async def broken_fake_summ_ai_server(request):
    """
    Create fake responses which simulate the SUMM.AI API server

    :param request: The request
    :type request: aiohttp.web.Request

    :return: The response
    :rtype: aiohttp.web.Response
    """
    return web.json_response(
        data={
            "error": "An error occurred",
        },
        status=500,
    )


# pylint: disable=unused-argument,too-many-locals
@pytest.mark.django_db
async def test_summ_ai_error_handling(
    login_role_user_async, settings, aiohttp_raw_server
):
    """
    This test checks whether the error handling of the SUMM.AI API client works as expected

    :param login_role_user_async: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user_async: tuple

    :param settings: The fixture providing the django settings
    :type settings: :fixture:`settings`

    :param aiohttp_raw_server: The fixture providing the dummy aiohttp server used for faking the SUMM.AI API server
    :type aiohttp_raw_server: :data:`aiohttp:pytest_aiohttp.aiohttp_raw_server`
    """
    # The region we want to use for testing
    region_slug = "augsburg"
    # Setup a dummy server to fake responses from the SUMM.AI API server (an instance of aiohttp.test_utils.RawTestServer)
    fake_server = await aiohttp_raw_server(broken_fake_summ_ai_server)
    # Enable SUMM.AI in the test region
    await enable_summ_api(region_slug)
    # Redirect call to the SUMM.AI API to the fake server
    settings.SUMM_AI_API_URL = (
        f"{fake_server.scheme}://{fake_server.host}:{fake_server.port}"
    )
    # Test for english messages
    settings.LANGUAGE_CODE = "en"
    # Log the user in
    client, role = login_role_user_async
    # Translate the pages
    translate_easy_german = reverse(
        "auto_translate_easy_german_pages",
        kwargs={"region_slug": region_slug, "language_slug": "de"},
    )
    response = await client.post(translate_easy_german, data={"selected_ids[]": [1]})
    print(response.headers)
    if role in PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]:
        # If the role should be allowed to access the view, we expect a redirect to the page tree
        assert response.status_code == 302
        page_tree = reverse(
            "pages", kwargs={"region_slug": region_slug, "language_slug": "de"}
        )
        assert response.headers.get("Location") == page_tree
        response = await client.get(page_tree)
        print(response.headers)
        # Check that the error message is present
        assert (
            "could not be automatically translated into Easy German."
            in response.content.decode("utf-8")
        )



"""
Django URL dispatcher for the cms package.
See :mod:`~integreat_cms.core.urls` for the other namespaces of this application.

For more information on this file, see :doc:`django:topics/http/urls`.
"""

from django.conf.urls import include, url


#: The url patterns of this module (see :doc:`django:topics/http/urls`)
urlpatterns = [
    url(r"^", include("integreat_cms.cms.urls.public")),
    url(r"^", include("integreat_cms.cms.urls.protected")),
]



import logging
import deepl

from django.conf import settings
from django.contrib import messages
from django.utils.translation import ugettext as _
from django.apps import apps

logger = logging.getLogger(__name__)


class DeepLApi:
    """
    DeepL API to auto translate selected posts.
    """

    def __init__(self):
        """
        Initialize the DeepL client
        """
        self.translator = deepl.Translator(settings.DEEPL_AUTH_KEY)

    @staticmethod
    def check_availability(request, language_slug):
        """
        This function checks, if the selected language is supported by DeepL

        :param request: request that was sent to the server
        :type request: ~django.http.HttpRequest

        :param language_slug: current language slug
        :type language_slug: str

        :return: true or false
        :rtype: bool
        """
        deepl_config = apps.get_app_config("deepl_api")
        source_language = request.region.get_source_language(language_slug)
        return (
            source_language
            and source_language.slug in deepl_config.supported_source_languages
            and language_slug in deepl_config.supported_target_languages
        )

    def deepl_translation(self, request, content_objects, language_slug, form_class):
        """
        This functions gets the translation from DeepL

        :param request: passed request
        :type request: ~django.http.HttpRequest

        :param content_objects: passed content objects
        :type content_objects: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]

        :param language_slug: current GUI language slug
        :type language_slug: str

        :param form_class: passed Form class of content type
        :type form_class: ~integreat_cms.cms.forms.custom_content_model_form.CustomContentModelForm
        """
        # Get target language
        target_language = request.region.get_language_or_404(language_slug)
        source_language = request.region.get_source_language(language_slug)
        for content_object in content_objects:
            source_translation = content_object.get_translation(source_language.slug)
            if not source_translation:
                messages.error(
                    request,
                    _('No source translation could be found for {} "{}".').format(
                        type(content_object)._meta.verbose_name.title(),
                        content_object.best_translation.title,
                    ),
                )
                continue
            existing_target_translation = content_object.get_translation(
                target_language.slug
            )
            # For some languages, the DeepL client expects the BCP tag instead of the short language code
            if target_language.slug in ("en", "pt"):
                target_language_key = target_language.bcp47_tag
            else:
                target_language_key = target_language.slug
            target_title = self.translator.translate_text(
                source_translation.title,
                source_lang=source_language.slug,
                target_lang=target_language_key,
            )
            data = {
                "title": target_title,
                "status": existing_target_translation.status
                if existing_target_translation
                else source_translation.status,
            }
            if source_translation.content:
                data["content"] = self.translator.translate_text(
                    source_translation.content,
                    source_lang=source_language.slug,
                    target_lang=target_language_key,
                )
            # for pois adds a short description
            if hasattr(source_translation, "short_description"):
                data["short_description"] = self.translator.translate_text(
                    source_translation.short_description,
                    source_lang=source_language.slug,
                    target_lang=target_language_key,
                )
            content_translation_form = form_class(
                data=data,
                instance=existing_target_translation,
                additional_instance_attributes={
                    "creator": request.user,
                    "language": target_language,
                    source_translation.foreign_field(): content_object,
                },
            )
            # Validate event translation
            if content_translation_form.is_valid():
                content_translation_form.save()
                logger.debug(
                    "Successfully translated for: %r", content_translation_form.instance
                )
                messages.success(
                    request,
                    _('{} "{}" has been successfully translated.').format(
                        type(content_object)._meta.verbose_name.title(),
                        source_translation.title,
                    ),
                )
            else:
                logger.error(
                    "Automatic translation for %r could not be created because of %r",
                    content_object,
                    content_translation_form.errors,
                )
                messages.error(
                    request,
                    _('{} "{}" could not be automatically translated.').format(
                        type(content_object)._meta.verbose_name.title(),
                        source_translation.title,
                    ),
                )



"""
This is a collection of tags and filters which are useful for all user forms
"""
import logging

from django import template

logger = logging.getLogger(__name__)
register = template.Library()


@register.filter
def remaining_regions(user, region_to_remove):
    """
    This calculates the remaining regions of a user after a given region is removed

    :param user: The given user
    :type user: ~django.contrib.auth.models.User

    :param region_to_remove: The region which should be removed from the user's regions
    :type region_to_remove:  ~integreat_cms.cms.models.regions.region.Region

    :return: The Queryset of regions
    :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.regions.region.Region ]
    """
    return user.regions.exclude(slug=region_to_remove.slug)



"""
Expansion of API-Endpoints for the CMS
"""
from django.urls import include, path, re_path

from .v3.events import events
from .v3.feedback import (
    page_feedback,
    search_result_feedback,
    region_feedback,
    offer_feedback,
    offer_list_feedback,
    event_list_feedback,
    event_feedback,
    poi_feedback,
    map_feedback,
    imprint_page_feedback,
    legacy_feedback_endpoint,
)
from .v3.imprint import imprint
from .v3.languages import languages
from .v3.locations import locations
from .v3.pages import (
    pages,
    children,
    parents,
    single_page,
    push_page_translation_content,
)
from .v3.pdf_export import pdf_export
from .v3.push_notifications import sent_push_notifications
from .v3.regions import regions, liveregions, hiddenregions
from .v3.offers import offers


#: The namespace for this URL config (see :attr:`django.urls.ResolverMatch.app_name`)
app_name = "api"

content_api_urlpatterns = [
    path("pages/", pages, name="pages"),
    path("locations/", locations, name="locations"),
    path("events/", events, name="events"),
    path("page/", single_page, name="single_page"),
    path("post/", single_page, name="single_page"),
    path("children/", children, name="children"),
    path("parents/", parents, name="parents"),
    path("pdf/", pdf_export, name="pdf_export"),
    path(
        "fcm/",
        sent_push_notifications,
        name="sent_push_notifications",
    ),
    path("imprint/", imprint, name="imprint"),
    path("disclaimer/", imprint, name="imprint"),
    path("offers/", offers, name="offers"),
    path("extras/", offers, name="offers"),
    path(
        "pushpage/", push_page_translation_content, name="push_page_translation_content"
    ),
    re_path(
        r"^feedback/?$",
        legacy_feedback_endpoint.legacy_feedback_endpoint,
        name="legacy_feedback_endpoint",
    ),
    path(
        "feedback/",
        include(
            [
                re_path(
                    r"^categories/?$",
                    region_feedback.region_feedback,
                    name="region_feedback",
                ),
                re_path(r"^page/?$", page_feedback.page_feedback, name="page_feedback"),
                re_path(r"^poi/?$", poi_feedback.poi_feedback, name="poi_feedback"),
                re_path(
                    r"^event/?$", event_feedback.event_feedback, name="event_feedback"
                ),
                re_path(
                    r"^events/?$",
                    event_list_feedback.event_list_feedback,
                    name="event_list_feedback",
                ),
                re_path(
                    r"^imprint-page/?$",
                    imprint_page_feedback.imprint_page_feedback,
                    name="imprint_page_feedbacks",
                ),
                re_path(r"^map/?$", map_feedback.map_feedback, name="map_feedback"),
                re_path(
                    r"^search/?$",
                    search_result_feedback.search_result_feedback,
                    name="search_result_feedback",
                ),
                re_path(
                    r"^offers/?$",
                    offer_list_feedback.offer_list_feedback,
                    name="offer_list_feedback",
                ),
                re_path(
                    r"^extras/?$",
                    offer_list_feedback.offer_list_feedback,
                    name="offer_list_feedback",
                ),
                re_path(
                    r"^offer/?$", offer_feedback.offer_feedback, name="offer_feedback"
                ),
                re_path(
                    r"^extra/?$", offer_feedback.offer_feedback, name="offer_feedback"
                ),
            ]
        ),
    ),
]

region_api_urlpatterns = [
    path("", regions, name="regions"),
    path("live/", liveregions, name="regions_live"),
    path("hidden/", hiddenregions, name="regions_hidden"),
]

#: The url patterns of this module (see :doc:`django:topics/http/urls`)
urlpatterns = [
    path("api/regions/", include(region_api_urlpatterns)),
    path("wp-json/extensions/v3/sites/", include(region_api_urlpatterns)),
    path(
        "api/<slug:region_slug>/",
        include(
            [
                path("languages/", languages, name="languages"),
                path("offers/", offers, name="offers"),
                path("extras/", offers, name="offers"),
                path("<slug:language_slug>/", include(content_api_urlpatterns)),
            ]
        ),
    ),
    path(
        "<slug:region_slug>/",
        include(
            [
                path(
                    "de/wp-json/extensions/v3/languages/", languages, name="languages"
                ),
                path(
                    "<slug:language_slug>/wp-json/extensions/v3/",
                    include(content_api_urlpatterns),
                ),
                path(
                    "<slug:language_slug>/wp-json/ig-mpdf/v1/pdf/",
                    pdf_export,
                    name="pdf_export",
                ),
            ]
        ),
    ),
]



import logging

from django.shortcuts import reverse
from django.views.generic.base import RedirectView

logger = logging.getLogger(__name__)


class LinkcheckRedirectView(RedirectView):
    """
    View for redirecting to main page of the broken link checker
    """

    def get_redirect_url(self, *args, **kwargs):
        r"""
        Retrieve url for redirection

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: url to redirect to
        :rtype: str
        """
        kwargs.update({"url_filter": "invalid"})
        return reverse("linkcheck", kwargs=kwargs)



import logging

from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.shortcuts import render, redirect
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ...decorators import permission_required
from ...forms import ImprintTranslationForm
from ...models import Language, ImprintPage

logger = logging.getLogger(__name__)


@method_decorator(permission_required("cms.view_imprintpage"), name="dispatch")
@method_decorator(permission_required("cms.change_imprintpage"), name="post")
class ImprintSideBySideView(TemplateView):
    """
    View for the imprint side by side form
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "imprint/imprint_sbs.html"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {
        "current_menu_item": "imprint",
        "IMPRINT_SLUG": settings.IMPRINT_SLUG,
    }

    def get(self, request, *args, **kwargs):
        r"""
        Render :class:`~integreat_cms.cms.forms.imprint.imprint_translation_form.ImprintTranslationForm` on the side by side view

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: If no imprint exists for the region

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region

        try:
            imprint = region.imprint
        except ImprintPage.DoesNotExist as e:
            raise Http404 from e

        disabled = False
        # Make form disabled if user has no permission to manage the imprint
        if not request.user.has_perm("cms.change_imprintpage"):
            disabled = True
            messages.warning(
                request, _("You don't have the permission to edit the imprint.")
            )

        target_language = Language.objects.get(slug=kwargs.get("language_slug"))
        source_language_node = region.language_tree_nodes.get(
            language=target_language
        ).parent

        if source_language_node:
            source_language = source_language_node.language
        else:
            messages.error(
                request,
                _(
                    "You cannot use the side-by-side-view for the region's default language (in this case {default_language})."
                ).format(default_language=target_language.translated_name),
            )
            return redirect(
                "edit_imprint",
                **{
                    "region_slug": region.slug,
                    "language_slug": target_language.slug,
                }
            )

        source_imprint_translation = imprint.get_translation(source_language.slug)
        target_imprint_translation = imprint.get_translation(target_language.slug)

        if not source_imprint_translation:
            messages.error(
                request,
                _(
                    "You cannot use the side-by-side-view if the source translation (in this case {source_language}) does not exist."
                ).format(source_language=source_language.translated_name),
            )
            return redirect(
                "edit_imprint",
                **{
                    "region_slug": region.slug,
                    "language_slug": target_language.slug,
                }
            )

        imprint_translation_form = ImprintTranslationForm(
            instance=target_imprint_translation, disabled=disabled
        )

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "imprint_translation_form": imprint_translation_form,
                "source_imprint_translation": source_imprint_translation,
                "target_language": target_language,
            },
        )

    # pylint: disable=unused-argument, too-many-branches
    def post(self, request, *args, **kwargs):
        r"""
        Submit :class:`~integreat_cms.cms.forms.imprint.imprint_translation_form.ImprintTranslationForm` and save
        :class:`~integreat_cms.cms.models.pages.imprint_page_translation.ImprintPageTranslation` object

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :raises ~django.http.Http404: If no imprint exists for the region

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region

        try:
            imprint = region.imprint
        except ImprintPage.DoesNotExist as e:
            raise Http404 from e

        target_language = Language.objects.get(slug=kwargs.get("language_slug"))
        source_language_node = region.language_tree_nodes.get(
            language=target_language
        ).parent

        if source_language_node:
            source_imprint_translation = imprint.get_translation(
                source_language_node.language.slug
            )
        else:
            messages.error(
                request,
                _(
                    "You cannot use the side-by-side-view for the region's default language (in this case {default_language})."
                ).format(default_language=target_language.translated_name),
            )
            return redirect(
                "edit_imprint",
                **{
                    "region_slug": region.slug,
                    "language_slug": target_language.slug,
                }
            )

        imprint_translation_instance = imprint.get_translation(target_language.slug)

        if not source_imprint_translation:
            messages.error(
                request,
                _(
                    "You cannot use the side-by-side-view if the source translation (in this case {source_language}) does not exist."
                ).format(source_language=source_language_node.language.translated_name),
            )
            return redirect(
                "edit_imprint",
                **{
                    "region_slug": region.slug,
                    "language_slug": target_language.slug,
                }
            )

        imprint_translation_form = ImprintTranslationForm(
            data=request.POST,
            instance=imprint_translation_instance,
            additional_instance_attributes={
                "page": imprint,
                "creator": request.user,
                "language": target_language,
            },
        )

        if not imprint_translation_form.is_valid():
            # Add error messages
            imprint_translation_form.add_error_messages(request)
        elif not imprint_translation_form.has_changed():
            # Add "no changes" messages
            messages.info(request, _("No changes made"))
        else:
            # Save form
            imprint_translation_form.save()
            # Add the success message
            imprint_translation_form.add_success_message(request)

        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "imprint_translation_form": imprint_translation_form,
                "source_imprint_translation": source_imprint_translation,
                "target_language": target_language,
            },
        )



from django import forms
from django.urls import reverse


class ParentFieldWidget(forms.widgets.Select):
    """
    This Widget class is used to append the url for retrieving the page order tables to the data attributes of the options
    """

    #: The form this field is bound to
    form = None

    # pylint: disable=too-many-arguments
    def create_option(
        self, name, value, label, selected, index, subindex=None, attrs=None
    ):
        """
        This function creates an option which can be selected in the parent field

        :param name: The name of the option
        :type name: str

        :param value: the value of the option (the page id)
        :type value: int

        :param label: The label of the option
        :type label: str

        :param selected: Whether or not the option is selected
        :type selected: bool

        :param index: The index of the option
        :type index: int

        :param subindex: The subindex of the option
        :type subindex: int

        :param attrs: The attributes of the option
        :type attrs: dict

        :return: The option dict
        :rtype: dict
        """
        # Create dictionary of options
        option_dict = super().create_option(
            name, value, label, selected, index, subindex=subindex, attrs=attrs
        )
        kwargs = {"region_slug": self.form.instance.region.slug}
        if value:
            kwargs["parent_id"] = value
        if self.form.instance.id:
            kwargs["page_id"] = self.form.instance.id
        option_dict["attrs"]["data-url"] = reverse(
            "get_page_order_table_ajax",
            kwargs=kwargs,
        )
        return option_dict



"""
APIv3 legacy feedback endpoint for pages, events and imprint
"""
from django.conf import settings
from django.http import JsonResponse

from .event_feedback import event_feedback_internal
from .imprint_page_feedback import imprint_page_feedback_internal
from .page_feedback import page_feedback_internal

from ...decorators import json_response, feedback_handler


@feedback_handler
@json_response
def legacy_feedback_endpoint(data, region, language, comment, rating, is_technical):
    """
    Decorate function for storing feedback about single page, imprint or event in database. This
    is a legacy endpoint for compatibility.

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :return: decorated function that saves feedback in database
    :rtype: ~collections.abc.Callable
    """
    link = data.get("permalink")
    if not link:
        return JsonResponse({"error": "Link is required."}, status=400)
    link_components = list(filter(None, link.split("/")))
    if link_components[-1] == settings.IMPRINT_SLUG:
        return imprint_page_feedback_internal(
            data, region, language, comment, rating, is_technical
        )
    data["slug"] = link_components[-1]
    if link_components[-2] == "events":
        return event_feedback_internal(
            data, region, language, comment, rating, is_technical
        )
    return page_feedback_internal(data, region, language, comment, rating, is_technical)



"""
This module contains utils for the sitemap app.
"""
import logging

from .sitemaps import PageSitemap, EventSitemap, POISitemap, OfferSitemap


logger = logging.getLogger(__name__)


def get_sitemaps(region, language):
    """
    This helper function generates a list of all non-empty sitemaps for a given region and language
    It is used in :class:`~integreat_cms.sitemap.views.SitemapIndexView` and :class:`~integreat_cms.sitemap.views.SitemapView`.

    :param region: The requested region
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The requested language
    :type language: ~integreat_cms.cms.models.languages.language.Language


    :return: All sitemaps for the given region and language
    :rtype: list [ ~sitemap.sitemaps.WebappSitemap ]
    """
    sitemaps = [
        PageSitemap(region, language),
        POISitemap(region, language),
        OfferSitemap(region, language),
    ]
    if region.events_enabled:
        sitemaps.append(EventSitemap(region, language))

    # Only return sitemaps which actually contain items
    sitemaps = [sitemap for sitemap in sitemaps if sitemap.items().exists()]

    logger.debug("Sitemaps for %r and %r: %r", region, language, sitemaps)

    return sitemaps



"""
This module contains all string representations of event time range filter options, used by
:class:`~integreat_cms.cms.forms.events.event_filter_form.EventFilterForm` and
:class:`~integreat_cms.cms.views.events.event_list_view.EventListView`.
"""
from django.utils.translation import ugettext_lazy as _


#: Events in the future
UPCOMING = "UPCOMING"
#: Events in the past
PAST = "PAST"
#: Events in a custom time range
CUSTOM = "CUSTOM"

#: Choices which indicate that no filtering is required
ALL_EVENTS = [UPCOMING, PAST]

#: Choices to use these constants in a form field
CHOICES = (
    (CUSTOM, _("Custom time range")),
    (UPCOMING, _("Upcoming events")),
    (PAST, _("Past events")),
)



"""
This module includes functions related to the push notification that are sent via firebase.
"""
from django.conf import settings
from django.http import JsonResponse
from django.utils import timezone

from ...cms.models import PushNotificationTranslation
from ..decorators import json_response


@json_response
def sent_push_notifications(request, region_slug, language_slug):
    """
    Function to iterate through all sent push notifications related to a region and adds them to a JSON.

    :param request: Django request
    :type request: ~django.http.HttpRequest
    :param region_slug: slug of a region
    :type region_slug: str
    :param language_slug: language slug
    :type language_slug: str

    :return: JSON object according to APIv3 push notifications definition
    :rtype: ~django.http.JsonResponse
    """
    channel = request.GET.get("channel", "all")
    query_result = (
        PushNotificationTranslation.objects.filter(
            push_notification__region__slug=region_slug
        )
        .filter(
            push_notification__sent_date__gte=timezone.now()
            - timezone.timedelta(days=settings.FCM_HISTORY_DAYS)
        )
        .filter(language__slug=language_slug)
        .order_by("-last_updated")
    )
    if channel != "all":
        query_result = query_result.filter(push_notification__channel=channel)
    result = list(map(transform_notification, query_result))
    return JsonResponse(result, safe=False)


def transform_notification(pnt):
    """
    Function to create a JSON from a single push notification translation Object.

    :param pnt: A push notification translation
    :type pnt: ~integreat_cms.cms.models.push_notifications.push_notification_translation.PushNotificationTranslation

    :return: data necessary for API
    :rtype: dict
    """
    return {
        "id": str(pnt.pk),
        "title": pnt.title,
        "message": pnt.text,
        "timestamp": pnt.last_updated,  # deprecated field in the future
        "last_updated": timezone.localtime(pnt.last_updated),
        "channel": pnt.push_notification.channel,
    }



"""
This module includes functions related to the event API endpoint.
"""

from datetime import datetime, timedelta

from django.conf import settings
from django.http import JsonResponse
from django.utils import timezone
from django.utils.html import strip_tags
from ..decorators import json_response
from .locations import transform_poi


def transform_event(event, custom_date=None):
    """
    Function to create a JSON from a single event object.

    :param event: The event which should be converted
    :type event: ~integreat_cms.cms.models.events.event.Event

    :param custom_date: The date overwrite of the event
    :type custom_date: ~datetime.date

    :return: data necessary for API
    :rtype: dict
    """
    start_local = event.start_local
    end_local = event.end_local
    if custom_date:
        start_local = datetime.combine(
            custom_date, start_local.time(), tzinfo=start_local.tzinfo
        )
        end_local = datetime.combine(
            custom_date + (end_local.date() - start_local.date()),
            end_local.time(),
            tzinfo=end_local.tzinfo,
        )

    return {
        "id": event.id if not custom_date else None,
        "start": start_local,
        "start_date": start_local.date(),  # deprecated field in the future
        "start_time": start_local.time(),  # deprecated field in the future
        "end": end_local,
        "end_date": end_local.date(),  # deprecated field in the future
        "end_time": end_local.time(),  # deprecated field in the future
        "all_day": event.is_all_day,
        "recurrence_id": event.recurrence_rule.id if event.recurrence_rule else None,
        "timezone": event.timezone,
    }


def transform_event_translation(event_translation, recurrence_date=None):
    """
    Function to create a JSON from a single event_translation object.

    :param event_translation: The event translation object which should be converted
    :type event_translation: ~integreat_cms.cms.models.events.event_translation.EventTranslation

    :param recurrence_date: The recurrence date for the event
    :type recurrence_date: ~datetime.date

    :return: data necessary for API
    :rtype: dict
    """
    event = event_translation.event
    slug = (
        event_translation.slug
        if not recurrence_date
        else f"{event_translation.slug}${recurrence_date}"
    )
    absolute_url = event_translation.url_prefix + slug + "/"
    return {
        "id": event_translation.id,
        "url": settings.BASE_URL + absolute_url,
        "path": absolute_url,
        "title": event_translation.title,
        "modified_gmt": event_translation.last_updated,  # deprecated field in the future
        "last_updated": timezone.localtime(event_translation.last_updated),
        "excerpt": strip_tags(event_translation.content),
        "content": event_translation.content,
        "available_languages": transform_available_languages(
            event_translation, recurrence_date
        )
        if recurrence_date
        else event_translation.available_languages_dict,
        "thumbnail": event.icon.url if event.icon else None,
        "location": transform_poi(event.location),
        "event": transform_event(event, recurrence_date),
        "hash": None,
    }


def transform_available_languages(event_translation, recurrence_date):
    """
    Function to create a JSON object of all available translations of an event translation.
    This is similar to `event_translation.available_languages_dict` embeds the recurrence date in the translation slug.

    :param event_translation: The event translation object which should be converted
    :type event_translation: ~integreat_cms.cms.models.events.event_translation.EventTranslation

    :param recurrence_date: The date of this event translation
    :type recurrence_date: ~datetime.date

    :return: data necessary for API
    :rtype: dict
    """
    languages = {}

    for translation in event_translation.foreign_object.available_translations():
        if translation.language == event_translation.language:
            continue

        slug = f"{translation.slug}${recurrence_date}"
        path = translation.url_prefix + slug + "/"
        languages[translation.language.slug] = {
            "id": None,
            "url": settings.BASE_URL + path,
            "path": path,
        }

    return languages


def transform_event_recurrences(event_translation, today):
    """
    Yield all future recurrences of the event.

    :param event_translation: The event translation object which should be converted
    :type event_translation: ~integreat_cms.cms.models.events.event_translation.EventTranslation

    :param today: The first date at which event may be yielded
    :type today: ~datetime.date

    :return: An iterator over all future recurrences up to ``settings.API_EVENTS_MAX_TIME_SPAN_DAYS``
    :rtype: Iterator[:class:`~datetime.date`]
    """
    event = event_translation.event
    recurrence_rule = event.recurrence_rule
    if not recurrence_rule:
        return

    # In order to avoid unnecessary computations, check if any future event
    # may be valid and return early if that is not the case
    if (
        recurrence_rule.recurrence_end_date
        and recurrence_rule.recurrence_end_date < today
    ):
        return

    start_date = event.start_local.date()
    event_translation.id = None

    # Calculate all recurrences of this event
    for recurrence_date in recurrence_rule.iter_after(start_date):
        if recurrence_date - max(start_date, today) > timedelta(
            days=settings.API_EVENTS_MAX_TIME_SPAN_DAYS
        ):
            break
        if recurrence_date < today or recurrence_date == start_date:
            continue

        yield transform_event_translation(event_translation, recurrence_date)


@json_response
# pylint: disable=unused-argument
def events(request, region_slug, language_slug):
    """
    List all events of the region and transform result into JSON

    :param request: The current request
    :type request: ~django.http.HttpRequest

    :param region_slug: The slug of the requested region
    :type region_slug: str

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: JSON object according to APIv3 events endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    region = request.region
    # Throw a 404 error when the language does not exist or is disabled
    region.get_language_or_404(language_slug, only_active=True)
    result = []
    now = timezone.now().date()
    for event in region.events.prefetch_public_translations().filter(archived=False):
        event_translation = event.get_public_translation(language_slug)
        if event_translation:
            if event.end_local.date() >= now:
                result.append(transform_event_translation(event_translation))

            for future_event in transform_event_recurrences(event_translation, now):
                result.append(future_event)

    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays



from django.db import models
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from ..pages.page_translation import PageTranslation
from .feedback import Feedback


class PageFeedback(Feedback):
    """
    Database model representing feedback about pages.
    """

    page_translation = models.ForeignKey(
        PageTranslation,
        on_delete=models.CASCADE,
        related_name="feedback",
        verbose_name=_("page translation"),
    )

    @property
    def object_name(self):
        """
        This property returns the name of the object this feedback comments on.

        :return: The name of the object this feedback refers to
        :rtype: str
        """
        return self.best_page_translation.title

    @cached_property
    def object_url(self):
        """
        This property returns the url to the object this feedback comments on.

        :return: The url to the referred object
        :rtype: str
        """
        return reverse(
            "edit_page",
            kwargs={
                "page_id": self.page_translation.page.id,
                "region_slug": self.region.slug,
                "language_slug": self.best_page_translation.language.slug,
            },
        )

    @cached_property
    def best_page_translation(self):
        """
        This property returns the best translation for the page this feedback comments on.

        :return: The best page translation
        :rtype: ~integreat_cms.cms.models.pages.page_translation.PageTranslation
        """
        return self.page_translation.page.best_translation

    @property
    def related_feedback(self):
        """
        This property returns all feedback entries which relate to the same object and have the same is_technical value.

        :return: The queryset of related feedback
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.feedback.page_feedback.PageFeedback ]
        """
        return PageFeedback.objects.filter(
            page_translation__page=self.page_translation.page,
            language=self.language,
            is_technical=self.is_technical,
        )

    class Meta:
        #: The verbose name of the model
        verbose_name = _("page feedback")
        #: The plural verbose name of the model
        verbose_name_plural = _("page feedback")
        #: The default permissions for this model
        default_permissions = ()



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from cacheops import invalidate_obj

from ..models import (
    Feedback,
    EventFeedback,
    EventListFeedback,
    PageFeedback,
    POIFeedback,
    ImprintPageFeedback,
    MapFeedback,
    OfferFeedback,
    OfferListFeedback,
    RegionFeedback,
    SearchResultFeedback,
)


@receiver(post_delete, sender=Feedback)
# pylint: disable=unused-argument
def feedback_delete_handler(sender, **kwargs):
    r"""
    Invalidate feedback cache after feedback deletion

    :param sender: The class of the feedback that was deleted
    :type sender: type

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict
    """
    if kwargs.get("instance"):
        invalidate_obj(kwargs.get("instance"))


@receiver(post_save, sender=PageFeedback)
@receiver(post_save, sender=EventFeedback)
@receiver(post_save, sender=EventListFeedback)
@receiver(post_save, sender=POIFeedback)
@receiver(post_save, sender=ImprintPageFeedback)
@receiver(post_save, sender=MapFeedback)
@receiver(post_save, sender=OfferFeedback)
@receiver(post_save, sender=OfferListFeedback)
@receiver(post_save, sender=RegionFeedback)
@receiver(post_save, sender=SearchResultFeedback)
# pylint: disable=unused-argument
def feedback_create_handler(sender, **kwargs):
    r"""
    Invalidate feedback cache after feedback creation

    :param sender: The class of the feedback that was deleted
    :type sender: type

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict
    """
    if kwargs.get("instance") and hasattr(kwargs.get("instance"), "feedback_ptr"):
        invalidate_obj(kwargs.get("instance").feedback_ptr)



"""
This module contains helpers for the translation process.
"""
from django.utils.text import format_lazy


def ugettext_many_lazy(*strings):
    r"""
    This function is a wrapper for :func:`django.utils.text.format_lazy` for the special case that the given strings
    should be concatenated with a space in between. This is useful for splitting lazy translated strings by sentences
    which improves the translation memory.

    :param \*strings: A list of lazy translated strings which should be concatenated
    :type \*strings: list

    :return: A lazy formatted string
    :rtype: str
    """
    fstring = ("{} " * len(strings)).strip()
    return format_lazy(fstring, *strings)



"""
This module contains view actions for objects related to POIs.
"""
import logging

from django.contrib import messages
from django.http import Http404
from django.shortcuts import render, redirect
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from linkcheck.models import Link

from ...decorators import permission_required
from ...models import POI

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.change_poi")
def archive_poi(request, poi_id, region_slug, language_slug):
    """
    Archive POI object

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param poi_id: The id of the POI which should be archived
    :type poi_id: int

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param language_slug: The slug of the current language
    :type language_slug: str

    :return: A redirection to the :class:`~integreat_cms.cms.views.pois.poi_list_view.POIListView`
    :rtype: ~django.http.HttpResponseRedirect
    """
    poi = POI.objects.get(id=poi_id)

    poi.archived = True
    poi.save()

    # Delete related link objects as they are no longer required
    Link.objects.filter(poi_translation__poi=poi).delete()

    logger.debug("%r archived by %r", poi, request.user)
    messages.success(request, _("Location was successfully archived"))

    return redirect(
        "pois",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@require_POST
@permission_required("cms.change_poi")
def restore_poi(request, poi_id, region_slug, language_slug):
    """
    Restore POI object (set ``archived=False``)

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param poi_id: The id of the POI which should be restored
    :type poi_id: int

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param language_slug: The slug of the current language
    :type language_slug: str

    :return: A redirection to the :class:`~integreat_cms.cms.views.pois.poi_list_view.POIListView`
    :rtype: ~django.http.HttpResponseRedirect
    """
    poi = POI.objects.get(id=poi_id)

    poi.archived = False
    poi.save()

    # Restore related link objects
    for translation in poi.translations.distinct("poi__pk", "language__pk"):
        # The post_save signal will create link objects from the content
        translation.save(update_timestamp=False)

    logger.debug("%r restored by %r", poi, request.user)
    messages.success(request, _("Location was successfully restored"))

    return redirect(
        "pois",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@require_POST
@permission_required("cms.delete_poi")
def delete_poi(request, poi_id, region_slug, language_slug):
    """
    Delete POI object

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param poi_id: The id of the POI which should be deleted
    :type poi_id: int

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param language_slug: The slug of the current language
    :type language_slug: str

    :return: A redirection to the :class:`~integreat_cms.cms.views.pois.poi_list_view.POIListView`
    :rtype: ~django.http.HttpResponseRedirect
    """

    poi = POI.objects.get(id=poi_id)
    logger.debug("%r deleted by %r", poi, request.user)
    poi.delete()
    messages.success(request, _("Location was successfully deleted"))

    return redirect(
        "pois",
        **{
            "region_slug": region_slug,
            "language_slug": language_slug,
        }
    )


@permission_required("cms.view_poi")
# pylint: disable=unused-argument
def view_poi(request, poi_id, region_slug, language_slug):
    """
    View POI object

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param poi_id: The id of the POI which should be viewed
    :type poi_id: int

    :param region_slug: The slug of the current region
    :type region_slug: str

    :param language_slug: The slug of the current language
    :type language_slug: str

    :raises ~django.http.Http404: If user no translation exists for the requested POI and language

    :return: A redirection to the :class:`~integreat_cms.cms.views.pois.poi_list_view.POIListView`
    :rtype: ~django.http.HttpResponseRedirect
    """
    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "pois/poi_view.html"
    poi = POI.objects.get(id=poi_id)

    poi_translation = poi.get_translation(language_slug)

    if not poi_translation:
        raise Http404

    return render(request, template_name, {"poi_translation": poi_translation})



import pytest

from django.contrib import auth
from django.urls import reverse


# pylint: disable=unused-argument
@pytest.mark.django_db
@pytest.mark.parametrize(
    "username", ["root", "root@root.root", "management", "management@example.com"]
)
def test_login_success(load_test_data, client, settings, username):
    """
    Test whether login via username & email works as expected

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple

    :param client: The fixture providing the an unauthenticated user client
    :type client: :fixture:`client`

    :param settings: The Django settings
    :type settings: :fixture:`settings`

    :param username: The username/email to use for login
    :type username: str
    """
    # Test login via username/password
    response = client.post(
        settings.LOGIN_URL, data={"username": username, "password": "root1234"}
    )
    print(response.headers)
    assert response.status_code == 302
    assert response.headers.get("location") == settings.LOGIN_REDIRECT_URL
    response = client.get(settings.LOGIN_REDIRECT_URL)
    print(response.headers)
    assert response.status_code == 302
    user = auth.get_user(client)
    assert user.is_authenticated
    if user.is_superuser or user.is_staff:
        # Root user should get redirected to the admin dashboard
        assert response.headers.get("location") == reverse("admin_dashboard")
    else:
        # Region user should get redirected to the region dashboard
        assert response.headers.get("location") == reverse(
            "dashboard", kwargs={"region_slug": "augsburg"}
        )


# pylint: disable=unused-argument
@pytest.mark.django_db
@pytest.mark.parametrize(
    "username",
    [
        "root",
        "root@root.root",
        "management",
        "management@example.com",
        "non-existing-user",
        "non-existing-email@example.com",
        "",
    ],
)
def test_login_failure(load_test_data, client, settings, username):
    """
    Test whether login with incorrect credentials does not work

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple

    :param client: The fixture providing the an unauthenticated user client
    :type client: :fixture:`client`

    :param settings: The Django settings
    :type settings: :fixture:`settings`

    :param username: The username/email to use for login
    :type username: str
    """
    # Test for english messages
    settings.LANGUAGE_CODE = "en"
    # Test login via username/password
    response = client.post(
        settings.LOGIN_URL, data={"username": username, "password": "incorrect"}
    )
    print(response.headers)
    assert response.status_code == 200
    assert "The username or the password is incorrect." in response.content.decode()



"""
Module for sending Push Notifications
"""
import logging
import requests

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from ...models import PushNotificationTranslation
from ...models import Region
from ...constants import push_notifications as pnt_const

logger = logging.getLogger(__name__)


class PushNotificationSender:
    """
    Sends push notifications via FCM HTTP API.
    Definition: https://firebase.google.com/docs/cloud-messaging/http-server-ref#downstream-http-messages-json
    """

    fcm_url = "https://fcm.googleapis.com/fcm/send"

    def __init__(self, push_notification):
        """
        Load relevant push notification translations and prepare content for sending

        :param push_notification: the push notification that should be sent
        :type push_notification: ~integreat_cms.cms.models.push_notifications.push_notification.PushNotification

        :raises ~django.core.exceptions.ImproperlyConfigured: If the auth key is missing or the system runs in debug
                                                              mode but the test region does not exist.
        """
        self.push_notification = push_notification
        self.prepared_pnts = []
        self.primary_pnt = PushNotificationTranslation.objects.get(
            push_notification=push_notification,
            language=push_notification.region.default_language,
        )
        if self.primary_pnt.title:
            self.prepared_pnts.append(self.primary_pnt)
        self.load_secondary_pnts()

        if not settings.FCM_ENABLED:
            raise ImproperlyConfigured("Push notifications are disabled")
        self.auth_key = settings.FCM_KEY

        if settings.DEBUG:
            # Prevent sending PNs to actual users in development
            test_region = Region.objects.filter(slug=settings.TEST_REGION_SLUG).first()
            if not test_region:
                raise ImproperlyConfigured(
                    f"The system runs with DEBUG=True but the region with TEST_REGION_SLUG={settings.TEST_REGION_SLUG} does not exist."
                )
            self.region = test_region
        self.region = push_notification.region

    def load_secondary_pnts(self):
        """
        Load push notification translations in other languages
        """
        secondary_pnts = PushNotificationTranslation.objects.filter(
            push_notification=self.push_notification
        ).exclude(id=self.primary_pnt.id)
        for secondary_pnt in secondary_pnts:
            if (
                secondary_pnt.title == ""
                and pnt_const.USE_MAIN_LANGUAGE == self.push_notification.mode
            ):
                secondary_pnt.title = self.primary_pnt.title
                secondary_pnt.text = self.primary_pnt.text
                self.prepared_pnts.append(secondary_pnt)
            if len(secondary_pnt.title) > 0:
                self.prepared_pnts.append(secondary_pnt)

    def is_valid(self):
        """
        Check if all data for sending push notifications is available

        :return: all prepared push notification translations are valid
        :rtype: bool
        """
        if not self.prepared_pnts:
            logger.debug(
                "%r does not have a default translation", self.push_notification
            )
            return False
        for pnt in self.prepared_pnts:
            if not pnt.title:
                logger.debug("%r has no title", pnt)
                return False
        return True

    def send_pn(self, pnt):
        """
        Send single push notification translation

        :param pnt: the prepared push notification translation to be sent
        :type pnt: ~integreat_cms.cms.models.push_notifications.push_notification_translation.PushNotificationTranslation

        :return: Response of the :mod:`requests` library
        :rtype: ~requests.Response
        """
        payload = {
            "to": f"/topics/{self.region.slug}-{pnt.language.slug}-{self.push_notification.channel}",
            "notification": {"title": pnt.title, "body": pnt.text},
            "data": {
                "news_id": str(pnt.id),
                "city_code": self.region.slug,
                "language_code": pnt.language.slug,
                "group": self.push_notification.channel,
            },
            "apns": {
                "headers": {"apns-priority": "5"},
            },
            "android": {
                "ttl": "86400s",
            },
            "payload": {
                "aps": {
                    "category": "NEW_MESSAGE_CATEGORY",
                }
            },
        }
        headers = {"Authorization": f"key={self.auth_key}"}
        return requests.post(
            self.fcm_url,
            json=payload,
            headers=headers,
            timeout=settings.DEFAULT_REQUEST_TIMEOUT,
        )

    def send_all(self):
        """
        Send all prepared push notification translations

        :return: Success status
        :rtype: bool
        """
        status = True
        for pnt in self.prepared_pnts:
            res = self.send_pn(pnt)
            if res.status_code == 200:
                logger.info("%r sent, FCM id: %r", pnt, res.json()["message_id"])
            else:
                status = False
                logger.error(
                    "Received invalid response from FCM for %r, status: %r, body: %r",
                    pnt,
                    res.status_code,
                    res.text,
                )
        return status



import logging

from django.views.generic import TemplateView

from ...models import Feedback
from ..chat.chat_context_mixin import ChatContextMixin

logger = logging.getLogger(__name__)


class AdminDashboardView(TemplateView, ChatContextMixin):
    """
    View for the admin dashboard
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "dashboard/admin_dashboard.html"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {"current_menu_item": "admin_dashboard"}

    def get_context_data(self, **kwargs):
        r"""
        Returns a dictionary representing the template context
        (see :meth:`~django.views.generic.base.ContextMixin.get_context_data`).

        :param \**kwargs: The given keyword arguments
        :type \**kwargs: dict

        :return: The template context
        :rtype: dict
        """
        context = super().get_context_data(**kwargs)
        context["admin_feedback"] = Feedback.objects.filter(
            is_technical=True, read_by=None
        )[:5]
        return context



"""
This package contains views for the broken link checker
"""
from .linkcheck_list_view import LinkcheckListView
from .linkcheck_redirect_view import LinkcheckRedirectView
from .linkcheck_stats_view import LinkcheckStatsView



import logging

from ...models import LanguageTreeNode
from ..list_views import ModelListView
from .language_tree_context_mixin import LanguageTreeContextMixin

logger = logging.getLogger(__name__)


# pylint: disable=too-many-ancestors
class LanguageTreeView(LanguageTreeContextMixin, ModelListView):
    """
    View for rendering the language tree view.
    This view is available in regions.
    """

    #: The model of this list view
    model = LanguageTreeNode
    #: Disable pagination for language tree
    paginate_by = None

    def get_queryset(self):
        """
        Get language tree queryset

        :return: The language tree of the current region
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.languages.language_tree_node.LanguageTreeNode ]
        """
        # Return the annotated language tree of the current region to save a few database queries
        return self.request.region.language_tree



from ...models import Organization
from ..icon_widget import IconWidget
from ..custom_model_form import CustomModelForm


class OrganizationForm(CustomModelForm):
    """
    Form for creating and modifying organization objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = Organization
        #: The fields of the model which should be handled by this form
        fields = [
            "name",
            "slug",
            "icon",
        ]
        #: The widgets which are used in this form
        widgets = {
            "icon": IconWidget(),
        }



"""
Configuration of Nominatim API app
"""
import os
import sys
import logging

from django.apps import AppConfig
from django.conf import settings

from .nominatim_api_client import NominatimApiClient

logger = logging.getLogger(__name__)


class NominatimApiConfig(AppConfig):
    """
    Nominatim API config inheriting the Django AppConfig
    """

    name = "integreat_cms.nominatim_api"

    def ready(self):
        """
        Checking if API is available
        """
        # Only check availability if running a server
        if "runserver" in sys.argv or "APACHE_PID_FILE" in os.environ:
            # If Nominatim API is enabled, check availability
            if settings.NOMINATIM_API_ENABLED:
                NominatimApiClient().check_availability()
            else:
                logger.info("Nominatim API is disabled")



"""
This module contains view actions for region objects.
"""
import logging

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from ...decorators import permission_required
from ...models import Region

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.delete_region")
# pylint: disable=unused-argument
def delete_region(request, *args, **kwargs):
    r"""
    This view deletes a region. All content is cascade deleted. Region users, who are not assigned to any other region,
    are manually removed.

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param \*args: The supplied arguments
    :type \*args: list

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict

    :return: A redirection to the media library
    :rtype: ~django.http.HttpResponseRedirect
    """

    region = get_object_or_404(Region, slug=kwargs.get("slug"))
    # Remove hierarchy to prevent ProtectedError when children get deleted before their parents
    region.pages.update(parent=None)
    region.language_tree_nodes.update(parent=None)
    # Prevent ProtectedError when location gets deleted before their events
    region.events.update(location=None)
    # Delete region and cascade delete all contents
    deleted_objects = region.delete()
    logger.info(
        "%r deleted %r, cascade deleted objects: %r",
        request.user,
        region,
        deleted_objects,
    )
    # Get orphan users who aren't superuser or staff and don't have a region assigned
    # (Creating users with these combination is impossible, so they were region users of the deleted region before)
    orphan_users = get_user_model().objects.filter(
        is_superuser=False, is_staff=False, regions=None
    )
    if orphan_users.exists():
        logger.info(
            "Deleted orphan users: %r",
            orphan_users,
        )
        orphan_users.delete()

    messages.success(request, _("Region was successfully deleted"))

    return redirect("regions")



"""
This module contains the possible status of translations.
"""
from django.utils.translation import ugettext_lazy as _


#: Draft
DRAFT = "DRAFT"
#: Pending Review
REVIEW = "REVIEW"
#: Public
PUBLIC = "PUBLIC"
#: Auto Save
AUTO_SAVE = "AUTO_SAVE"

#: Choices to use these constants in a database field
CHOICES = (
    (DRAFT, _("Draft")),
    (REVIEW, _("Pending Review")),
    (PUBLIC, _("Published")),
    (AUTO_SAVE, _("Auto Save")),
)



"""
This package contains all views related to settings
"""
from .user_settings_view import UserSettingsView
from .dismiss_tutorial_view import DismissTutorial

from .mfa.authenticate_modify_mfa_view import AuthenticateModifyMfaView
from .mfa.delete_user_mfa_key_view import DeleteUserMfaKeyView
from .mfa.get_mfa_challenge_view import GetMfaChallengeView
from .mfa.register_user_mfa_key_view import RegisterUserMfaKeyView



"""
This package contains all views related to feedback
"""
from .admin_feedback_list_view import AdminFeedbackListView
from .region_feedback_list_view import RegionFeedbackListView
from .admin_feedback_actions import (
    mark_admin_feedback_as_read,
    mark_admin_feedback_as_unread,
    delete_admin_feedback,
)
from .region_feedback_actions import (
    mark_region_feedback_as_read,
    mark_region_feedback_as_unread,
    delete_region_feedback,
)



from django.conf import settings
from django.urls import reverse

from ...conftest import ANONYMOUS


def check_view_status_code(login_role_user, view_name, kwargs, post_data, roles):
    """
    This test checks whether the given view return the correct status code for the current role

    :param login_role_user: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user: tuple

    :param view_name: The identifier of the view
    :type view_name: str

    :param kwargs: The keyword argument passed to the view
    :type kwargs: dict

    :param post_data: The post data for this view
    :type post_data: dict

    :param roles: The list of roles which should be able to access this view
    :type roles: list
    """
    client, role = login_role_user
    url = reverse(view_name, kwargs=kwargs)
    if post_data:
        kwargs = {"data": post_data}
        # If the post data is a string, assume json as content type
        if isinstance(post_data, str):
            kwargs["content_type"] = "application/json"
        response = client.post(url, **kwargs)
    else:
        response = client.get(url)
    print(response.headers)
    if role in roles:
        # If the role should be allowed to access the view, we expect a successful result
        if post_data:
            if response.headers.get("Content-Type") == "application/json":
                # Json-views should return 200 or 201 CREATED (for feedback)
                assert response.status_code in [200, 201]
            else:
                # Normal post-views should redirect after a successful operation (200 usually mean form errors)
                assert response.status_code == 302
        else:
            # Get-views should return 200
            assert response.status_code == 200
    elif role == ANONYMOUS:
        # For anonymous users, we want to redirect to the login form instead of showing an error
        assert response.status_code == 302
        assert response.headers.get("location") == f"{settings.LOGIN_URL}?next={url}"
    else:
        # For logged in users, we want to show an error if they get a permission denied
        assert response.status_code == 403



"""
This module contains all string representations of all day filter options, used by :class:`~integreat_cms.cms.forms.events.event_filter_form.EventFilterForm` and
:class:`~integreat_cms.cms.views.events.event_list_view.EventListView`.

The module also contains a constant :attr:`~integreat_cms.cms.constants.all_day.DATATYPE`, which contains the type of the constant
values linked to the strings and is used for correctly instantiating :class:`django.forms.TypedMultipleChoiceField`
instances in :class:`~integreat_cms.cms.forms.events.event_filter_form.EventFilterForm`.
"""
from django.utils.translation import ugettext_lazy as _


#: This docstring is overridden by "alias of builtins.int"
DATATYPE = int

#: Only events which are all day long
ALL_DAY = 1
#: Exclude events which are all day long
NOT_ALL_DAY = 2

#: Choices to use these constants in a database field
CHOICES = (
    (ALL_DAY, _("All day long events")),
    (NOT_ALL_DAY, _("Not all day long events")),
)



"""
Meta information about the Integreat CMS API app
"""
from django.apps import AppConfig


class ApiConfig(AppConfig):
    """
    Configuration parameters for the API app
    """

    name = "integreat_cms.api"



from django.db import models
from django.core.validators import MaxValueValidator, MinValueValidator
from django.utils.translation import ugettext_lazy as _

from ..abstract_content_model import AbstractContentModel
from ..media.media_file import MediaFile
from ..pois.poi_translation import POITranslation


class POI(AbstractContentModel):
    """
    Data model representing a point of interest (POI). It contains all relevant data about its exact position, including
    coordinates.
    """

    address = models.CharField(
        max_length=250, verbose_name=_("street and house number")
    )
    postcode = models.CharField(max_length=10, verbose_name=_("postal code"))
    city = models.CharField(max_length=250, verbose_name=_("city"))
    country = models.CharField(max_length=250, verbose_name=_("country"))
    latitude = models.FloatField(
        null=True,
        blank=True,
        verbose_name=_("latitude"),
        help_text=_("The latitude coordinate"),
        validators=[MinValueValidator(-90.0), MaxValueValidator(90.0)],
    )
    longitude = models.FloatField(
        null=True,
        blank=True,
        verbose_name=_("longitude"),
        help_text=_("The longitude coordinate"),
        validators=[MinValueValidator(-180.0), MaxValueValidator(180.0)],
    )
    location_on_map = models.BooleanField(
        default=False,
        verbose_name=_("Show this location on map"),
        help_text=_("Tick if you want to show this location on map"),
    )
    icon = models.ForeignKey(
        MediaFile,
        verbose_name=_("icon"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    archived = models.BooleanField(
        default=False,
        verbose_name=_("archived"),
        help_text=_("Whether or not the location is read-only and hidden in the API."),
    )
    website = models.URLField(max_length=250, blank=True, verbose_name=_("website"))
    email = models.EmailField(
        blank=True,
        verbose_name=_("email address"),
    )
    phone_number = models.CharField(
        max_length=250, blank=True, verbose_name=_("phone number")
    )

    @property
    def fallback_translations_enabled(self):
        """
        Whether translations should be returned in the default language if they do not exist

        :return: Whether fallback translations are enabled
        :rtype: bool
        """
        return self.region.fallback_translations_enabled

    @staticmethod
    def get_translation_model():
        """
        Returns the translation model of this content model

        :return: The class of translations
        :rtype: type
        """
        return POITranslation

    class Meta:
        #: The verbose name of the model
        verbose_name = _("location")
        #: The plural verbose name of the model
        verbose_name_plural = _("locations")
        #: The name that will be used by default for the relation from a related object back to this one
        default_related_name = "pois"
        #: The default permissions for this model
        default_permissions = ("change", "delete", "view")
        #: The fields which are used to sort the returned objects of a QuerySet
        ordering = ["pk"]



import logging

from django import forms
from django.conf import settings
from django.utils.translation import gettext as _

from geopy.distance import distance

from ....nominatim_api.nominatim_api_client import NominatimApiClient
from ...models import POI
from ..custom_model_form import CustomModelForm
from ..icon_widget import IconWidget


logger = logging.getLogger(__name__)


class POIForm(CustomModelForm):
    """
    Form for creating and modifying POI objects
    """

    #: The distance in km between the manually entered coordinates and the coordinates returned from Nominatim
    nominatim_distance_delta = 0

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = POI
        #: The fields of the model which should be handled by this form
        fields = [
            "address",
            "postcode",
            "city",
            "country",
            "latitude",
            "longitude",
            "location_on_map",
            "icon",
            "website",
            "email",
            "phone_number",
        ]
        #: The widgets which are used in this form
        widgets = {
            "icon": IconWidget(),
        }

    def clean(self):
        """
        Validate form fields which depend on each other, see :meth:`django.forms.Form.clean`

        :return: The cleaned form data
        :rtype: dict
        """
        cleaned_data = super().clean()

        # When the Nominatim API is enabled, validate the coordinates
        if settings.NOMINATIM_API_ENABLED:
            nominatim_api_client = NominatimApiClient()
            latitude, longitude = nominatim_api_client.get_coordinates(
                street=cleaned_data.get("address"),
                postalcode=cleaned_data.get("postcode"),
                city=cleaned_data.get("city"),
            )
            if latitude and longitude:
                # Only override coordinates if not set manually
                if not cleaned_data.get("latitude"):
                    cleaned_data["latitude"] = latitude
                if not cleaned_data.get("longitude"):
                    cleaned_data["longitude"] = longitude
                # Store distance between manually entered coordinates and API result
                self.nominatim_distance_delta = int(
                    distance(
                        (cleaned_data["latitude"], cleaned_data["longitude"]),
                        (latitude, longitude),
                    ).km
                )

        if cleaned_data.get("location_on_map"):
            # If the location should be shown on the map, require the coordinates
            if not cleaned_data.get("latitude"):
                self.add_error(
                    "latitude",
                    forms.ValidationError(
                        _(
                            "Could not derive the coordinates from the address, please fill "
                            "the field manually if the location is to be displayed on the map."
                        ),
                        code="required",
                    ),
                )
            if not cleaned_data.get("longitude"):
                self.add_error(
                    "longitude",
                    forms.ValidationError(
                        _(
                            "Could not derive the coordinates from the address, please fill "
                            "the field manually if the location is to be displayed on the map."
                        ),
                        code="required",
                    ),
                )

        return cleaned_data



from django import template

register = template.Library()


@register.filter
def in_list(value, comma_separated_list):
    """
    This filter checks whether a given string value is contained in a comma separated list

    :param value: The value which should be checked
    :type value: str

    :param comma_separated_list: The list of comma-separated strings
    :type comma_separated_list: str

    :return: Whether or not value is contained in the list
    :rtype: bool
    """
    return value in comma_separated_list.split(",")


@register.filter
def get_private_member(element, key):
    """
    This filter returns a private member of an element

    :param element: The requested object
    :type element: object

    :param key: The key of the private variable (without the leading underscore)
    :type key: str

    :return: The value of the private variable
    :rtype: object
    """
    return element[f"_{key}"]



import logging

from django.http import JsonResponse, Http404

from ....cms.models import EventFeedback
from ...decorators import json_response, feedback_handler

logger = logging.getLogger(__name__)


@feedback_handler
@json_response
def event_feedback(data, region, language, comment, rating, is_technical):
    """
    Store feedback about single event in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :return: decorated function that saves feedback in database
    :rtype: ~collections.abc.Callable
    """
    return event_feedback_internal(
        data, region, language, comment, rating, is_technical
    )


def event_feedback_internal(data, region, language, comment, rating, is_technical):
    """
    Store feedback about single event in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :raises ~django.http.Http404: HTTP status 404 if no event with the given slug exists.

    :return: JSON object according to APIv3 single page feedback endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    event_translation_slug = data.get("slug")

    events = region.events.filter(
        translations__slug=data.get("slug"),
        translations__language=language,
    ).distinct()

    if len(events) > 1:
        logger.error(
            "Event translation slug %r is not unique per region and language, found multiple: %r",
            event_translation_slug,
            events,
        )
        return JsonResponse({"error": "Internal Server Error"}, status=500)
    if len(events) == 0:
        raise Http404("No matching event found for slug.")
    event = events[0]
    event_translation = event.get_translation(language.slug)

    EventFeedback.objects.create(
        event_translation=event_translation,
        region=region,
        language=language,
        rating=rating,
        comment=comment,
        is_technical=is_technical,
    )
    return JsonResponse({"success": "Feedback successfully submitted"}, status=201)



"""
This module includes functions related to the pages API endpoint.
"""
import logging
import json
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned
from django.http import JsonResponse, Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.utils.html import strip_tags
from django.utils.text import slugify
from django.utils import timezone
from ...cms.models import Page
from ...cms.forms import PageTranslationForm
from ..decorators import json_response, matomo_tracking

logger = logging.getLogger(__name__)


def transform_page(page_translation):
    """
    Function to create a dict from a single page_translation Object.

    :param page_translation: single page translation object
    :type page_translation: ~integreat_cms.cms.models.pages.page_translation.PageTranslation

    :raises ~django.http.Http404: HTTP status 404 if a parent is archived

    :return: data necessary for API
    :rtype: dict
    """
    fallback_parent = {
        "id": 0,
        "url": None,
        "path": None,
    }

    parent_page = page_translation.page.cached_parent
    if parent_page and not parent_page.explicitly_archived:
        parent_public_translation = parent_page.get_public_translation(
            page_translation.language.slug
        )
        if parent_public_translation:
            parent_absolute_url = parent_public_translation.get_absolute_url()
            parent = {
                "id": parent_page.id,
                "url": settings.BASE_URL + parent_absolute_url,
                "path": parent_absolute_url,
            }
            # use left edge indicator of mptt model for ordering of child pages
            order = page_translation.page.lft
        else:
            logger.info(
                "The parent %r of %r does not have a public translation in %r",
                parent_page,
                page_translation.page,
                page_translation.language,
            )
            raise Http404("No Page matches the given url or id.")
    else:
        parent = fallback_parent
        # use tree id of mptt model for ordering of root pages
        order = page_translation.page.tree_id

    absolute_url = page_translation.get_absolute_url()
    return {
        "id": page_translation.id,
        "url": settings.BASE_URL + absolute_url,
        "path": absolute_url,
        "title": page_translation.title,
        "modified_gmt": page_translation.combined_last_updated,  # deprecated field in the future
        "last_updated": timezone.localtime(page_translation.combined_last_updated),
        "excerpt": strip_tags(page_translation.combined_text),
        "content": page_translation.combined_text,
        "parent": parent,
        "order": order,
        "available_languages": page_translation.available_languages_dict,
        "thumbnail": page_translation.page.icon.url
        if page_translation.page.icon
        else None,
        "hash": None,
    }


@matomo_tracking
@json_response
# pylint: disable=unused-argument
def pages(request, region_slug, language_slug):
    """
    Function to iterate through all non-archived pages of a region and return them as JSON.

    :param request: Django request
    :type request: ~django.http.HttpRequest
    :param region_slug: slug of a region
    :type region_slug: str
    :param language_slug: language slug
    :type language_slug: str

    :return: JSON object according to APIv3 pages endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    region = request.region
    # Throw a 404 error when the language does not exist or is disabled
    region.get_language_or_404(language_slug, only_active=True)
    result = []
    # The preliminary filter for explicitly_archived=False is not strictly required, but reduces the number of entries
    # requested from the database
    for page in region.pages.filter(explicitly_archived=False).cache_tree(
        archived=False, language_slug=language_slug
    ):
        page_translation = page.get_public_translation(language_slug)
        if page_translation:
            result.append(transform_page(page_translation))
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays


def get_single_page(request, language_slug):
    """
    Helper function returning the desired page or a 404 if the
    requested page does not exist or is archived.

    :param request: The request that has been sent to the Django server
    :type request: ~django.http.HttpRequest

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.core.exceptions.MultipleObjectsReturned: If the given url cannot be resolved unambiguously

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :raises RuntimeError: If neither the id nor the url parameter is given

    :return: the requested page
    :rtype: ~integreat_cms.cms.models.pages.page.Page
    """
    region = request.region

    if request.GET.get("id"):
        page = get_object_or_404(region.pages, id=request.GET.get("id"))

    elif request.GET.get("url"):
        # Strip leading and trailing slashes to avoid ambiguous urls
        url = request.GET.get("url").strip("/")
        # The last path component of the url is the page translation slug
        page_translation_slug = slugify(url.split("/")[-1], allow_unicode=True)
        # Get page by filtering for translation slug and translation language slug
        filtered_pages = region.pages.filter(
            translations__slug=page_translation_slug,
            translations__language__slug=language_slug,
        ).distinct()

        if len(filtered_pages) > 1:
            logger.error(
                "Page translation slug %r is not unique per region and language, found multiple: %r",
                page_translation_slug,
                filtered_pages,
            )
            raise MultipleObjectsReturned(
                "This page translation slug is not unique, please contact your server administrator."
            )
        if len(filtered_pages) == 0:
            raise Http404("No matching page translation found for url.")
        page = filtered_pages[0]

    else:
        raise RuntimeError("Either the id or the url parameter is required.")

    if page.explicitly_archived:
        raise Http404("No matching page translation found for url.")

    # Check if any ancestor is archived -> will raise a 404
    get_public_ancestor_translations(page, language_slug)

    return page


@matomo_tracking
@json_response
# pylint: disable=unused-argument
def single_page(request, region_slug, language_slug):
    """
    View function returning the desired page as a JSON or a 404 if the
    requested page does not exist.

    :param request: The request that has been sent to the Django server
    :type request: ~django.http.HttpRequest

    :param region_slug: Slug defining the region
    :type region_slug: str

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :return: JSON with the requested page and a HTTP status 200.
    :rtype: ~django.http.JsonResponse
    """
    try:
        page = get_single_page(request, language_slug)
    except RuntimeError as e:
        return JsonResponse({"error": str(e)}, status=400)
    # Get most recent public revision of the page
    page_translation = page.get_public_translation(language_slug)
    if page_translation:
        return JsonResponse(transform_page(page_translation), safe=False)

    raise Http404("No Page matches the given url or id.")


@matomo_tracking
@json_response
# pylint: disable=unused-argument
def children(request, region_slug, language_slug):
    """
    Retrieves all children for a single page

    :param request: The request that has been sent to the Django server
    :type request: ~django.http.HttpRequest

    :param region_slug: Slug defining the region
    :type region_slug: str

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :return: JSON with the requested page descendants
    :rtype: ~django.http.JsonResponse
    """

    depth = int(request.GET.get("depth", 1))
    try:
        # try to get a single ancestor page based on the requests query string
        root_pages = [get_single_page(request, language_slug)]
    except RuntimeError:
        # if neither id nor url is set then get all root pages
        root_pages = Page.get_root_pages(region_slug)
        # simulate a virtual root node for WP compatibility
        # so that depth = 1 returns only those pages without parents (immediate children of this virtual root page)
        # like in wordpress depth = 0 will return no results in this case
        depth = depth - 1
    result = []
    public_region_pages = request.region.pages.filter(
        explicitly_archived=False, tree_id__in=[page.tree_id for page in root_pages]
    ).cache_tree(archived=False, language_slug=language_slug)
    for root in root_pages:
        descendants = root.get_tree_max_depth(max_depth=depth)
        for descendant in descendants:
            if descendant in public_region_pages:
                result.append(
                    transform_page(descendant.get_public_translation(language_slug))
                )
    return JsonResponse(result, safe=False)


@json_response
# pylint: disable=unused-argument
def parents(request, region_slug, language_slug):
    """
    Retrieves all ancestors (parent and all nodes up to the root node) of a page.
    If any ancestor is archived, an 404 is raised.

    :param request: The request that has been sent to the Django server
    :type request: ~django.http.HttpRequest

    :param region_slug: Slug defining the region
    :type region_slug: str

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :return: JSON with the requested page ancestors
    :rtype: ~django.http.JsonResponse
    """
    try:
        current_page = get_single_page(request, language_slug)
    except RuntimeError as e:
        return JsonResponse({"error": str(e)}, status=400)
    if not current_page.get_public_translation(language_slug):
        raise Http404("No Page matches the given url or id.")
    result = get_public_ancestor_translations(current_page, language_slug)
    return JsonResponse(result, safe=False)


def get_public_ancestor_translations(current_page, language_slug):
    """
    Retrieves all ancestors (parent and all nodes up to the root node) of a page.
    If any ancestor is archived or has a missing translation, a 404 is raised.

    :param current_page: the page that needs a list of its ancestor translations
    :type current_page: ~integreat_cms.cms.models.pages.page.Page

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :return: JSON with the requested page ancestors
    :rtype: ~django.http.JsonResponse
    """
    result = []
    for ancestor in current_page.get_cached_ancestors():
        public_translation = ancestor.get_public_translation(language_slug)
        if not public_translation or ancestor.explicitly_archived:
            raise Http404("No Page matches the given url or id.")
        result.append(transform_page(public_translation))
    return result


@csrf_exempt
@json_response
# pylint: disable=unused-argument
def push_page_translation_content(request, region_slug, language_slug):
    """
    Retrieves all ancestors (parent and all nodes up to the root node) of a page

    :param request: The request that has been sent to the Django server
    :type request: ~django.http.HttpRequest

    :param region_slug: Slug defining the region
    :type region_slug: str

    :param language_slug: Code to identify the desired language
    :type language_slug: str

    :raises ~django.http.Http404: HTTP status 404 if the request is malformed or no page with the given id or url exists.

    :return: JSON with the requested page ancestors
    :rtype: ~django.http.JsonResponse
    """
    try:
        data = json.loads(request.body)
    except json.decoder.JSONDecodeError as e:
        logger.error("Push Content: failed to parse JSON: %s", e)
        return JsonResponse({"status": "error"}, status=405)

    if not all(key in data for key in ["content", "token"]):
        logger.error("Push Content: missing required key.")
        return JsonResponse({"status": "error"}, status=405)

    page = request.region.pages.filter(api_token=data["token"]).first()
    if not page or data["token"] == "":
        return JsonResponse(data={"status": "denied"}, status=403)

    translation = page.get_translation(language_slug)
    data = {
        "content": data["content"],
        "title": translation.title,
        "slug": translation.slug,
        "status": translation.status,
        "minor_edit": False,
    }
    page_translation_form = PageTranslationForm(
        data=data,
        instance=translation,
        additional_instance_attributes={
            "page": page,
            "creator": None,
        },
    )
    if page_translation_form.is_valid():
        page_translation_form.save()
        return JsonResponse({"status": "success"})
    logger.error("Push Content: failed to validate page translation.")
    return JsonResponse({"status": "error"}, status=405)



from datetime import datetime
from time import mktime

from django import template

register = template.Library()


@register.filter
def parse_struct_time(struct_time):
    """
    Converts timestamp to datetime and returns date only

    :param struct_time: timestamp received from rss feed
    :type struct_time: time.struct_time
    :return: proper parsed date
    :rtype: datetime.date
    """
    return datetime.fromtimestamp(mktime(struct_time)).date()
